"""CHEBI API methods."""
from math import ceil

from flask import request

from ebel import Bel
from ebel.manager.rdbms.models import chebi
from ebel.web.api import RDBMS

from . import _get_pagination


def get_compound_by_name(name: str):
    """Get CHEBI compound information based given name.

    Parameters
    ----------
    name: str
        Compound name.

    Returns
    -------
    dict
        CHEBI compound information.
    """
    return [x.as_dict() for x in RDBMS.get_session().query(chebi.Compound).filter_by(name=name).all()]


def get_compound_name_by_name_starts_with(name: str):
    """Get CHEBI compound name based on fuzzy given value.

    Parameters
    ----------
    name: str
        Beginning of compound name.

    Returns
    -------
    dict
        Names and CHEBI IDs.
    """
    query = (
        RDBMS.get_session().query(chebi.Compound.id, chebi.Compound.name).filter(chebi.Compound.name.like(f"{name}%"))
    )
    return {x.name: x.id for x in query.all()}


def get_compound_by_id(id: int):
    """Get compound by CHEBI ID.

    Parameters
    ----------
    chebi_id: int
        CHEBI ID.

    Returns
    -------
    dict
        Compound information.
    """
    return RDBMS.get_session().query(chebi.Compound).filter_by(id=id).first().as_dict()


def get_compound_by_other_db_accession(accession_number: str, db_name: str = None):
    """Get a compound's information based on accession number.

    Parameters
    ----------
    accession_number: str
        Accession number.
    db_name: str
        Name of database where accession number was derived.

    Returns
    -------
    list
        List of dictionaries of compounds matching accession number.
    """
    query_filter = []
    query_filter.append(chebi.DatabaseAccession.accession_number == accession_number)
    if db_name:
        query_filter.append(chebi.DatabaseAccession.type == db_name)
    chebi_ids_rs = RDBMS.get_session().query(chebi.DatabaseAccession.compound_id).filter(*query_filter).all()
    chebi_ids = {x[0] for x in chebi_ids_rs}
    return [
        x.as_dict() for x in RDBMS.get_session().query(chebi.Compound).filter(chebi.Compound.id.in_(chebi_ids)).all()
    ]


def get_compound_reference():
    """Compile a dictionary of information for a given compound."""
    req = dict(request.args.copy())
    page_size = req.pop("page_size", 10)
    page = req.pop("page", 1)
    if not req:
        return {"error": "At least one of the parameters have to be filled."}
    query = RDBMS.get_session().query(chebi.Reference).filter_by(**req)
    number_of_results = query.count()

    limit = int(page_size)
    page = int(page)
    offset = (page - 1) * limit
    offset = offset if offset <= 100 else 100
    query = query.limit(limit).offset(offset)
    pages = ceil(number_of_results / limit)

    return {
        "page": page,
        "page_size": limit,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": [x.as_dict_with_compound_id() for x in query.all()],
    }


def get_relation():
    """Get CHEBI defined relations."""
    if not (bool(request.args.get("final_id") or bool(request.args.get("init_id")))):
        return {"error": "At least final_id or init_id have to be filled."}
    rs = RDBMS.get_session().query(chebi.Relation).filter_by(**request.args).all()
    return [x.as_dict() for x in rs]


def get_bel_chebi_ids():
    """Get BEL nodes with the CHEBI namespace."""
    b = Bel()
    sql_count = "SELECT count(*) FROM V where chebi IS NOT NULL"
    p = _get_pagination()
    number_of_results = b.query_get_dict(sql_count)[0]["count"]
    pages = ceil(number_of_results / p.page_size)
    query = "SELECT @rid.asString(), namespace, name, chebi, bel FROM V where chebi IS NOT NULL"
    paras = {k: v for k, v in dict(request.args.copy()).items() if k in ["namespace", "name", "chebi"]}
    if paras:
        query += " AND " + " AND ".join([f"{k} like '{v.strip()}'" for k, v in paras.items()])

    print(query)
    return {
        "page": p.page,
        "page_size": p.page_size,
        "number_of_results": number_of_results,
        "pages": pages,
        "results": b.query_get_dict(f"{query} LIMIT {p.page_size} SKIP {p.skip}"),
    }



"""ClinicalTrials.gov RDBMS model definition."""
import re

from sqlalchemy import Column, ForeignKey, Integer, String, Table, Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()

ctg_keyword_n2m = Table(
    "clinical_trials_gov__keyword",
    Base.metadata,
    Column(
        "clinical_trials_gov_keyword_id",
        Integer,
        ForeignKey("clinical_trials_gov_keyword.id"),
        index=True,
    ),
    Column(
        "clinical_trials_gov_id",
        Integer,
        ForeignKey("clinical_trials_gov.id"),
        index=True,
    ),
)

ctg_condition_n2m = Table(
    "clinical_trials_gov__condition",
    Base.metadata,
    Column(
        "clinical_trials_gov_condition_id",
        Integer,
        ForeignKey("clinical_trials_gov_condition.id"),
        index=True,
    ),
    Column(
        "clinical_trials_gov_id",
        Integer,
        ForeignKey("clinical_trials_gov.id"),
        index=True,
    ),
)

ctg_mesh_term_n2m = Table(
    "clinical_trials_gov__mesh_term",
    Base.metadata,
    Column(
        "clinical_trials_gov_mesh_term_id",
        Integer,
        ForeignKey("clinical_trials_gov_mesh_term.id"),
        index=True,
    ),
    Column(
        "clinical_trials_gov_id",
        Integer,
        ForeignKey("clinical_trials_gov.id"),
        index=True,
    ),
)

ctg_intervention_n2m = Table(
    "clinical_trials_gov__intervention",
    Base.metadata,
    Column(
        "clinical_trials_gov_intervention_id",
        Integer,
        ForeignKey("clinical_trials_gov_intervention.id"),
        index=True,
    ),
    Column(
        "clinical_trials_gov_id",
        Integer,
        ForeignKey("clinical_trials_gov.id"),
        index=True,
    ),
)


class ClinicalTrialGov(Base):
    """Class definition for the clinical_trials_gov table."""

    __tablename__ = "clinical_trials_gov"

    id = Column(Integer, primary_key=True)
    nct_id = Column(String(100), index=True)
    org_study_id = Column(Text)
    brief_title = Column(Text)
    official_title = Column(Text)
    is_fda_regulated_drug = Column(Text)
    brief_summary = Column(Text)
    detailed_description = Column(Text)
    overall_status = Column(Text)
    start_date = Column(Text)
    completion_date = Column(Text)
    phase = Column(Text)
    study_type = Column(Text)
    study_design_intervention_model = Column(Text)
    study_design_primary_purpose = Column(Text)
    study_design_masking = Column(Text)
    # primary_outcomes
    # secondary_outcomes
    patient_data_sharing_ipd = Column(Text)
    patient_data_ipd_description = Column(Text)

    keywords = relationship(
        "Keyword",
        secondary=ctg_keyword_n2m,
        back_populates="trials",
        cascade="save-update",
    )

    conditions = relationship(
        "Condition",
        secondary=ctg_condition_n2m,
        back_populates="trials",
        cascade="save-update",
    )

    mesh_terms = relationship(
        "MeshTerm",
        secondary=ctg_mesh_term_n2m,
        back_populates="trials",
        cascade="save-update",
    )

    interventions = relationship(
        "Intervention",
        secondary=ctg_intervention_n2m,
        back_populates="trials",
        cascade="save-update",
    )

    def as_dict(self):
        """Convert object values to dictionary."""
        basic_dict = object_as_dict(self)
        basic_dict["brief_summary"] = re.sub(r"\r?\n\s*", " ", basic_dict["brief_summary"]).strip()
        basic_dict.update({"keywords": [x.keyword for x in self.keywords]})
        basic_dict.update({"conditions": [x.condition for x in self.conditions]})
        basic_dict.update({"mesh_terms": [x.mesh_term for x in self.mesh_terms]})
        basic_dict.update(
            {
                "interventions": [
                    {
                        "intervention_type": x.intervention_type,
                        "intervention_name": x.intervention_name,
                    }
                    for x in self.interventions
                ]
            }
        )
        return basic_dict


class Keyword(Base):
    """Class definition for the clinical_trials_gov_keyword table."""

    __tablename__ = "clinical_trials_gov_keyword"
    id = Column(Integer, primary_key=True)
    keyword = Column(String(255), index=True)
    trials = relationship("ClinicalTrialGov", secondary=ctg_keyword_n2m, back_populates="keywords")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"keyword": self.keyword, "nct_ids": [x.nct_id for x in self.trials]}


class Condition(Base):
    """Class definition for the clinical_trials_gov_condition table."""

    __tablename__ = "clinical_trials_gov_condition"
    id = Column(Integer, primary_key=True)
    condition = Column(Text)
    trials = relationship("ClinicalTrialGov", secondary=ctg_condition_n2m, back_populates="conditions")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {"condition": self.condition, "nct_ids": [x.nct_id for x in self.trials]}


class MeshTerm(Base):
    """Class definition for the clinical_trials_gov_mesh_term table."""

    __tablename__ = "clinical_trials_gov_mesh_term"
    id = Column(Integer, primary_key=True)
    mesh_term = Column(String(100), unique=True)
    trials = relationship("ClinicalTrialGov", secondary=ctg_mesh_term_n2m, back_populates="mesh_terms")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "mesh_term": self.mesh_term,
            "number_of_trials": len(self.trials),
            "nct_ids": [x.nct_id for x in self.trials],
        }


class Intervention(Base):
    """Class definition for the clinical_trials_gov_intervention table."""

    __tablename__ = "clinical_trials_gov_intervention"
    id = Column(Integer, primary_key=True)
    intervention_type = Column(String(100), index=True)
    intervention_name = Column(String(255), index=True)
    trials = relationship(
        "ClinicalTrialGov",
        secondary=ctg_intervention_n2m,
        back_populates="interventions",
    )

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "intervention_type": self.intervention_type,
            "intervention_name": self.intervention_name,
            "nct_ids": [x.nct_id for x in self.trials],
        }



"""Pathway Commons API methods."""

from flask import request
from sqlalchemy import or_

from ebel.manager.rdbms.models.pathway_commons import (
    PathwayCommons, PathwayName, Pmid, pathway_commons__pathway_name)
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_data, _get_paginated_ebel_query_result,
                                  _get_paginated_query_result,
                                  _get_terms_from_model_starts_with)


def get_pathway_commons():
    """Get paginated list of Pathway Commons entries."""
    return _get_data(PathwayCommons)


def get_by_gene_symbol():
    """Get entries by gene symbol."""
    gene_symbol = request.args.get("gene_symbol")
    query = (
        RDBMS.get_session()
        .query(PathwayCommons)
        .filter(
            or_(
                PathwayCommons.participant_a == gene_symbol,
                PathwayCommons.participant_b == gene_symbol,
            )
        )
    )
    return _get_paginated_query_result(query)


def get_by_pathway_name():
    """Get entries by pathway name."""
    pathway_name = request.args.get("pathway_name")
    query = (
        RDBMS.get_session()
        .query(PathwayCommons)
        .join(pathway_commons__pathway_name)
        .join(PathwayName)
        .filter_by(name=pathway_name)
    )
    return _get_paginated_query_result(query)


def get_by_pmid():
    """Get entries by PMID."""
    pmid = request.args.get("pmid")
    query = RDBMS.get_session().query(PathwayCommons).join(Pmid).filter_by(pmid=pmid)
    return _get_paginated_query_result(query)


def get_pathway_name_starts_with():
    """Get entries where pathway starts with given value."""
    return _get_terms_from_model_starts_with("pathway_name", PathwayName.name)


def get_ebel_relation():
    """Get Pathway Commons related eBEL relations."""
    conf = {
        "interactor_a_rid": "out.rid.asString()",
        "interactor_a_namespace": "out.namespace",
        "interactor_a_name": "out.name",
        "edge_rid": "rid.asString()",
        "relation_type": "type",
        "sources": "sources",
        "pmids": "pmids",
        "pathways": "pathways.name",
        "interactor_b_rid": "in.rid.asString()",
        "interactor_b_namespace": "in.namespace",
        "interactor_b_name": "in.name",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_action_pc"

    ra = request.args
    wheres = []

    source = ra.get("source")
    if source:
        wheres.append(f'"{source}" in sources')

    pmid = ra.get("pmid")
    if source:
        wheres.append(f"{pmid} in pmids")

    pathway = ra.get("pathway")
    if pathway:
        wheres.append(f'"{pathway}" in pathways.name')

    paras = {k: request.args[k] for k in request.args if k in conf}
    if paras:
        wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]

    if wheres:
        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""ODB interface tests."""

from .constants import DB_NAME, PASSWORD, PORT, SERVER, USER, test_client

bel = test_client


class TestConfig:
    """Test class for configuration settings."""

    def test_connection_params(self):
        """Test that the proper connection parameters are being called and used during client initialization."""
        default_params = {"name": DB_NAME, "user": USER, "password": PASSWORD, "server": SERVER, "port": PORT}
        client_config = {
            "name": bel.odb_db_name,
            "user": bel.odb_user,
            "password": bel.odb_password,
            "server": bel.odb_server,
            "port": bel.odb_port,
        }

        # Check that the client configuration paramaters match the default ones initialized for testing
        assert all([client_config[param] == value for param, value in default_params.items()])



"""API server string constants."""
select_bel_columns_list = [
    "out.name as subject_name",
    "out.label as subject_label",
    "out.namespace as subject_namespace",
    "out.bel as subject_bel",
    "out.involved_genes as subject_involved_genes",
    "out.involved_other as subject_involved_other",
    "out.@class as subject_class",
    "out.@rid.asString() as subject_id",
    "@class as relation",
    "pmid",
    "@rid.asString() as edge_id",
    "evidence",
    "citation.last_author as last_author",
    "citation.title as title",
    "citation.pub_date as publication_date",
    "annotation",
    "in.name as object_name",
    "in.label as object_label",
    "in.namespace as object_namespace",
    "in.bel as object_bel",
    "in.involved_genes as object_involved_genes",
    "in.involved_other as object_involved_other",
    "in.@class as object_class",
    "in.@rid.asString() as object_id",
]

select_bel_columns = ", ".join(select_bel_columns_list)

match_bel_columns_list = [
    "s.bel as subject_bel",
    "s.involved_genes as subject_involved_genes",
    "s.involved_other as subject_involved_other",
    "s.@class as subject_class",
    "s.@rid.asString() as subject_id",
    "s.name as subject_name",
    "s.label as subject_label",
    "s.namespace as subject_namespace",
    "r.@class as relation",
    "r.pmid as pmid",
    "r.@rid.asString() as edge_id",
    "r.evidence as evidence",
    "r.citation.last_author as last_author",
    "r.citation.title as title",
    "r.citation.pub_date as publication_date",
    "r.annotation as annotation",
    "o.bel as object_bel",
    "o.involved_genes as object_involved_genes",
    "o.involved_other as object_involved_other",
    "o.@class as object_class",
    "o.@rid.asString() as object_id",
    "o.name as object_name",
    "o.label as object_label",
    "o.namespace as object_namespace",
]

match_bel_columns = ", ".join(match_bel_columns_list)



"""Command line interface to e(BE:L)."""

import logging
import sys

import click

import ebel.database
from ebel import Bel, web
from ebel.config import set_configuration, user_config_setup
from ebel.manager.orientdb.constants import DRUGBANK
from ebel.validate import repair_bel_file, validate_bel_file

logger = logging.getLogger(__name__)


@click.group(help="BEL test framework Command Line Utilities on {}".format(sys.executable))
@click.version_option()
def main():
    """Entry method."""
    pass


# TODO(@ceb): Implement by default "keep database" and new database as option #prio1
# TODO(@ceb): SQLAlchemy connection string #prio1
@main.command()
@click.argument("bel_script_path")
@click.option(
    "-l",
    "--line_by_line",
    is_flag=True,
    default=False,
    help="check script line by line",
)
@click.option(
    "-r",
    "--reports",
    default=None,
    help="path(s) to report file(s) seperated by comma with suffix (.md, .txt, .csv, .tsv, .json, .html)",
)
@click.option("-v", "--bel_version", default="2_1", help="stores a report in a file")
@click.option("-t", "--tree", is_flag=True, default=False, help="shows tree")
# TODO: Implement new Cytoscape export file
# @click.option('-c', '--cytoscape', is_flag=True, default=False, help='creates cytoscape file')
@click.option(
    "-s",
    "--sqlalchemy_connection_str",
    default=None,
    help="SQLAlchmy connection string",
)
@click.option("-j", "--json_file", is_flag=True, help="Create json file")
@click.option(
    "-f",
    "--force_json",
    is_flag=True,
    default=False,
    help="Force the creation of a JSON file",
)
def validate(
    bel_script_path: str,
    line_by_line: bool,
    reports: str,
    bel_version: str,
    tree: bool,
    sqlalchemy_connection_str: str,
    json_file: bool,
    force_json: bool,
):
    """Validate a BEL file using the defined grammar."""
    validate_bel_file(
        bel_script_path=bel_script_path,
        line_by_line=line_by_line,
        reports=reports,
        bel_version=bel_version,
        tree=tree,
        sqlalchemy_connection_str=sqlalchemy_connection_str,
        json_file=json_file,
        force_json=force_json,
    )


@main.command()
@click.argument("bel_script_path")
@click.option(
    "-n",
    "--new_file_path",
    default=None,
    help="Path to write repaired file to. If none passed, will overwrite original file.",
)
def repair(bel_script_path: str, new_file_path: str):
    """Repair the BEL file for common delimiters and line separations."""
    repair_bel_file(bel_script_path, new_file_path)


@main.command()
@click.argument("json_file_path")
@click.option(
    "-e",
    "--extend",
    is_flag=True,
    default=True,
    help='Flag to disable "extension" during import',
)
@click.option(
    "-g",
    "--p2g",
    is_flag=True,
    default=True,
    help='Flag to disable "protein2gene" during import',
)
@click.option(
    "-s",
    "--skip_drugbank",
    is_flag=True,
    default=False,
    help="Flag to disable DrugBank",
)
@click.option(
    "-i",
    "--include_subfolders",
    is_flag=True,
    default=False,
    help="Flag to enable directory walking",
)
@click.option("--drugbank_user", help="Valid username for DrugBank")
@click.option("--drugbank_password", help="Valid username for DrugBank")
@click.option("-n", "--odb_name", help="OrientDB database name")
@click.option("-u", "--odb_user", help="OrientDB user (with admin rights)")
@click.option("-p", "--odb_password", help="OrientDB user (with admin rights) password")
@click.option("-h", "--odb_server", help="OrientDB server name or URI")
@click.option("-o", "--odb_port", help="OrientDB server port")
@click.option("--odb_user_reader", help=" OrientDB user with only read rights")
@click.option("--odb_user_reader_password", help="OrientDB user with only read rights password")
@click.option(
    "--odb_root_password",
    help="OrientDB root user password (only during database setup)",
)
@click.option("--kegg_species", help="KEGG species")
@click.option("--sqlalchemy_connection_string", help="schema is user:password@server/database")
@click.option("--snp_related_traits", help="key of SNP related traits in GWAS catalog")
@click.option("--drugbank_user", help="Drugbank user")
@click.option("--drugbank_password", help="DrugBank password")
def import_json(
    json_file_path: str,
    extend: bool,
    p2g: bool,
    skip_drugbank: bool,
    drugbank_user: str,
    drugbank_password: str,
    include_subfolders: bool,
    odb_name: str,
    odb_user: str,
    odb_password: str,
    odb_server: str,
    odb_port: str,
    odb_user_reader: str,
    odb_user_reader_password: str,
    odb_root_password: str,
    kegg_species: str,
    sqlalchemy_connection_string: str,
    snp_related_traits: str,
):
    """Import JSON into OrientDB.

    Parameters
    ----------
    json_file_path : str
        Path to a BEL JSON file, list of BEL JSON files (separated by ","), or a directory containing BEL JSON files.
    extend : bool
        Boolean flag to disable network extension by e(BE:L)
    p2g : bool
        Boolean flag to disable adding genes and RNA to proteins in network
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    include_subfolders: bool (optional)
        Boolean flag to enable walking through subdirectories during import.
    odb_name: str
        OrientDB database name
    odb_user:
        OrientDB user (with admin rights)
    odb_password:
        OrientDB user (with admin rights) password
    odb_server:
        OrientDB server name or URI
    odb_port:
        OrientDB server port
    odb_user_reader:
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password:
        OrientDB root user password (only first database setup needed)
    kegg_species:
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits:
        SNP related traits

    Returns
    -------
    type
        returns True if imported.
    """
    # if one of the parameters is not None it will overwrite the default values from the configfile
    set_configuration(
        name=odb_name,
        user=odb_user,
        password=odb_password,
        server=odb_server,
        port=odb_port,
        user_reader=odb_user_reader,
        user_reader_password=odb_user_reader_password,
        root_password=odb_root_password,
        kegg_species=kegg_species,
        sqlalchemy_connection_string=sqlalchemy_connection_string,
        snp_related_traits=snp_related_traits,
        drugbank_user=drugbank_user,
        drugbank_password=drugbank_password,
    )

    bel = Bel()
    if "," in json_file_path:
        json_file_path = json_file_path.split()

    bel.import_json(
        input_path=json_file_path,
        extend_graph=extend,
        update_from_protein2gene=p2g,
        skip_drugbank=skip_drugbank,
        drugbank_user=drugbank_user,
        drugbank_password=drugbank_password,
        include_subfolders=include_subfolders,
    )


@main.command()
@click.option(
    "-s",
    "--skip",
    default=[],
    help="Comma-separated list of databases to skip during enrichment",
)
@click.option(
    "-i",
    "--include",
    default=[],
    help="Comma-separated list of databases to include during enrichment",
)
@click.option("--skip_drugbank", is_flag=True, default=False, help="Flag to disable DrugBank")
@click.option("--drugbank_user", default=None, help="Valid username for DrugBank")
@click.option("--drugbank_password", default=None, help="Valid username for DrugBank")
@click.option("-n", "--odb_name", default=None, help="OrientDB database name")
@click.option("-u", "--odb_user", default=None, help="OrientDB user (with admin rights)")
@click.option(
    "-p",
    "--odb_password",
    default=None,
    help="OrientDB user (with admin rights) password",
)
@click.option("-h", "--odb_server", default=None, help="OrientDB server name or URI")
@click.option("-o", "--odb_port", default=None, help="OrientDB server port")
@click.option("--odb_user_reader", default=None, help=" OrientDB user with only read rights")
@click.option(
    "--odb_user_reader_password",
    default=None,
    help="OrientDB user with only read rights password",
)
@click.option(
    "--odb_root_password",
    default=None,
    help="OrientDB root user password (only during database setup)",
)
@click.option("--kegg_species", default="hsa,rno,mmu", help="KEGG species")
@click.option("--sqlalchemy_connection_string", help="schema is user:password@server/database")
@click.option("--snp_related_traits", help="key of SNP related traits in GWAS catalog and ClinVar")
def enrich(
    skip: str,
    include: str,
    skip_drugbank: bool,
    drugbank_user: str,
    drugbank_password: str,
    odb_name: str,
    odb_user: str,
    odb_password: str,
    odb_server: str,
    odb_port: str,
    odb_user_reader: str,
    odb_user_reader_password: str,
    odb_root_password: str,
    kegg_species: str,
    sqlalchemy_connection_string: str,
    snp_related_traits: str,
):
    """Trigger the enrichment step for a database.

    Parameters
    ----------
    skip : str
        Comma-separated list of databases to skip during enrichment.
    include : str
        Comma-separated list of databases to include during enrichment.
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    odb_name: str
        OrientDB database name
    odb_user: str
        OrientDB user (with admin rights)
    odb_password: str
        OrientDB user (with admin rights) password
    odb_server: str
        OrientDB server name or URI
    odb_port: int
        OrientDB server port
    odb_user_reader: str
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password: str
        OrientDB root user password (only first database setup needed)
    kegg_species: str
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits: str
        SNP related traits
    """
    set_configuration(
        name=odb_name,
        user=odb_user,
        password=odb_password,
        server=odb_server,
        port=odb_port,
        user_reader=odb_user_reader,
        user_reader_password=odb_user_reader_password,
        root_password=odb_root_password,
        kegg_species=kegg_species,
        sqlalchemy_connection_string=sqlalchemy_connection_string,
        snp_related_traits=snp_related_traits,
        drugbank_user=drugbank_user,
        drugbank_password=drugbank_password,
    )

    bel = Bel()

    if not skip_drugbank:
        bel.drugbank.get_user_passwd(drugbank_user=drugbank_user, drugbank_password=drugbank_password)

    skip = skip.split(",") if isinstance(skip, str) else skip
    include = include.split(",") if isinstance(include, str) else include

    if skip_drugbank and DRUGBANK not in skip:
        skip.append(DRUGBANK)

    bel.enrich_network(skip=skip, include=include)


@main.command()
@click.argument("connection")
def set_connection(connection):
    """Set the SQLAlchemy connection string.

    :param connection: An existing RDBMS connection.
    """
    ebel.database.set_connection(connection)


@main.command()
@click.option("-h", "--host", default="localhost", help="MySQL server")
@click.option("-u", "--user", default="ebel_user", help="MySQL username")
@click.option("-p", "--password", default="ebel_passwd", help="MySQL password")
@click.option("-d", "--database", default="ebel", help="MySQL database name")
@click.option("-i", "--interactive", is_flag=True, default=False, help="Enable interactive mode")
def set_mysql(host: str, user: str, password: str, database: str, interactive: bool):
    """Set the SQLAlchemy connection string with MySQL settings."""
    if interactive:
        host, user, password, db = ebel.database.set_mysql_interactive()

    ebel.database.set_mysql_connection(host=host, user=user, password=password, db=database)


@main.command()
def settings():
    """Interactive method to create a configuration file."""
    user_config_setup()


@main.command()
@click.option("-h", "--host", default="0.0.0.0", help="Server or host name")
@click.option("-p", "--port", default="5000", help="server port [5000]")
@click.option("-d", "--debug_mode", is_flag=True, default=False, help="debug mode")
@click.option("-o", "--open_browser", is_flag=True, default=False, help="open browser")
def serve(host, port, debug_mode, open_browser):
    """Start the API RESTful server."""
    web.app.run(host=host, port=port, debug_mode=debug_mode, open_browser=open_browser)


if __name__ == "__main__":
    main()



"""URLs to download files."""

# HGNC #
HGNC_JSON = "ftp://ftp.ebi.ac.uk/pub/databases/genenames/new/json/hgnc_complete_set.json"
HGNC_TSV = "ftp://ftp.ebi.ac.uk/pub/databases/genenames/hgnc/tsv/hgnc_complete_set.txt"
HCOP_GZIP = "ftp://ftp.ebi.ac.uk/pub/databases/genenames/hcop/human_all_hcop_sixteen_column.txt.gz"

# UniProt #
UNIPROT_SPROT = (
    "ftp://ftp.uniprot.org/pub/databases/uniprot/" "current_release/knowledgebase/complete/uniprot_sprot.xml.gz"
)
UNIPROT_HGNC = (
    "https://www.genenames.org/cgi-bin/download/custom?col=gd_hgnc_id&col=gd_app_sym&col=md_prot_id&"
    "status=Approved&status=Entry%20Withdrawn&hgnc_dbtag=on&order_by=gd_app_sym_sort&"
    "format=text&submit=submit"
)
UNIPROT_MGI = "http://www.informatics.jax.org/downloads/reports/MRK_SwissProt.rpt"
UNIPROT_RGD = "https://download.rgd.mcw.edu/data_release/GENES_RAT.txt"
UNIPROT_FLYBASE = "ftp://ftp.flybase.org/releases/current/precomputed_files/genes/fbgn_NAseq_Uniprot_fb_2020_04.tsv.gz"

# Clinical Trials #
CLINICAL_TRIALS_GOV = "https://clinicaltrials.gov/AllPublicXML.zip"

# Genetic Variant DBs #
CLINVAR = "https://ftp.ncbi.nlm.nih.gov/pub/clinvar/tab_delimited/variant_summary.txt.gz"
# MIRTARBASE = "http://mirtarbase.cuhk.edu.cn/cache/download/8.0/miRTarBase_MTI.xlsx"
MIRTARBASE = "https://mirtarbase.cuhk.edu.cn/~miRTarBase/miRTarBase_2022/cache/download/9.0/miRTarBase_MTI.xlsx"
GWAS_CATALOG = "https://www.ebi.ac.uk/gwas/api/search/downloads/full"

# PPIs #
BIND = "https://www.bindingdb.org/bind/downloads/BindingDB_All_2018m7.tsv.zip"
BIOGRID = (
    "https://downloads.thebiogrid.org/Download/BioGRID/Release-Archive/BIOGRID-4.4.215/BIOGRID-ALL-4.4.215.tab3.zip"
)
INTACT = "ftp://ftp.ebi.ac.uk/pub/databases/intact/current/psimitab/intact.zip"
STITCH = "http://stitch.embl.de/download/protein_chemical.links.transfer.v5.0.tsv.gz"

# String #
STRING_INTS = "https://stringdb-static.org/download/protein.links.full.v11.5/9606.protein.links.full.v11.5.txt.gz"
STRING_ACTIONS = "https://stringdb-static.org/download/protein.actions.v11.0/9606.protein.actions.v11.0.txt.gz"
STRING_NAMES = "https://stringdb-static.org/download/protein.info.v11.5/9606.protein.info.v11.5.txt.gz"

# Pathway DBs #
KEGG_PATH_LIST = "http://rest.kegg.jp/list/pathway/hsa"
PATHWAY_COMMONS = "https://www.pathwaycommons.org/archives/PC2/v12/PathwayCommons12.Detailed.hgnc.txt.gz"
REACTOME = "https://reactome.org/download/current/UniProt2Reactome.txt"
# TODO: Import from Reactome MySQL
# REACTOME MySQL has a strange database structure and no controlled vocabulary, reactions are not classified
# For that reason I will work later on that
# https://reactome.org/content/schema/DatabaseObject
# REACTOME_MYSQL = "https://reactome.org/download/current/databases/gk_current.sql.gz"
WIKIPATHWAYS = "http://data.wikipathways.org/20180710/gpml/wikipathways-20180710-gpml-Homo_sapiens.zip"

# Ensembl #
ENSEMBL_FASTA_PEP = "ftp://ftp.ensembl.org/pub/release-94/fasta/homo_sapiens/pep/Homo_sapiens.GRCh38.pep.all.fa.gz"
ENSEMBL_CDS = "ftp://ftp.ensembl.org/pub/release-96/fasta/homo_sapiens/cds/Homo_sapiens.GRCh38.cds.all.fa.gz"

# SIDER #
SIDER_ATC = "http://sideeffects.embl.de/media/download/drug_atc.tsv"
SIDER_SE = "http://sideeffects.embl.de/media/download/meddra_all_se.tsv.gz"

# Expression Atlas #
EXPRESSION_ATLAS_BASE = "ftp://ftp.ebi.ac.uk/pub/databases/microarray/data/atlas/experiments/"
EXPRESSION_ATLAS_EXPERIMENTS = EXPRESSION_ATLAS_BASE + "atlas-latest-data.tar.gz"

# DisGeNet #
DISGENET_BASE = "https://www.disgenet.org/static/disgenet_ap1/files/downloads/"
DISGENET_GDP_ASSOC = DISGENET_BASE + "all_gene_disease_pmid_associations.tsv.gz"
DISGENET_VDP_ASSOC = DISGENET_BASE + "all_variant_disease_pmid_associations.tsv.gz"

# Drugs and Side Effects #
OFFSIDES = "http://tatonettilab.org/resources/nsides/OFFSIDES.csv.xz"
ONSIDES = "https://github.com/tatonetti-lab/onsides/releases/download/v2.0.0-20230629/onsides_v2.0.0_20230629.tar.gz"
# DrugBank
DRUGBANK_VERSION = "https://go.drugbank.ca/releases/latest#full"
DRUGBANK_DATA = "https://go.drugbank.com/releases/{}/downloads/all-full-database"
# IUPHAR
IUPHAR_INT = "https://www.guidetopharmacology.org/DATA/interactions.csv"
IUPHAR_LIGANDS = "https://www.guidetopharmacology.org/DATA/ligands.csv"

# CHEBI #
CHEBI_BASE = "ftp://ftp.ebi.ac.uk/pub/databases/chebi/Flat_file_tab_delimited/"
CHEBI_CHEMICALDATA = f"{CHEBI_BASE}chemical_data.tsv"
CHEBI_COMMENT = f"{CHEBI_BASE}comments.tsv"
CHEBI_COMPOUND = f"{CHEBI_BASE}compounds.tsv.gz"
CHEBI_DATABASEACCESSION = f"{CHEBI_BASE}database_accession.tsv"
CHEBI_NAME = f"{CHEBI_BASE}names.tsv.gz"
CHEBI_REFERENCE = f"{CHEBI_BASE}reference.tsv.gz"
CHEBI_RELATION = f"{CHEBI_BASE}relation.tsv"
CHEBI_STRUCTURE = f"{CHEBI_BASE}structures.csv.gz"
CHEBI_INCHI = f"{CHEBI_BASE}chebiId_inchi.tsv"
# CHEBI_COMPOUNDORIGIN = f"{CHEBI_BASE}compound_origins.tsv"

# NCBI #
NCBI_PMID = (
    "https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/?tool=ebel&" "ids={}&idtype=pmid&versions=no&format=json"
)
NCBI_MESH = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&retmode=xml&id="
# NCBI Gene
NCBI_GENE_INFO = "https://ftp.ncbi.nih.gov/gene/DATA/gene_info.gz"
NCBI_GENE_2ACCESSION = "https://ftp.ncbi.nih.gov/gene/DATA/gene2accession.gz"
NCBI_GENE_2ENSEMBL = "https://ftp.ncbi.nih.gov/gene/DATA/gene2ensembl.gz"
NCBI_GENE_2GO = "https://ftp.ncbi.nih.gov/gene/DATA/gene2go.gz"
NCBI_GENE_2MIM = "https://ftp.ncbi.nih.gov/gene/DATA/mim2gene_medgen"
NCBI_GENE_2PUBMED = "https://ftp.ncbi.nih.gov/gene/DATA/gene2pubmed.gz"
NCBI_GENE_ORTHOLOG = "https://ftp.ncbi.nih.gov/gene/DATA/gene_orthologs.gz"
NCBI_GENE_NEIGHBORS = "https://ftp.ncbi.nih.gov/gene/DATA/gene_neighbors.gz"
# NCBI MedGen
NCBI_GENE_MEDGEN_PUBMED = "https://ftp.ncbi.nlm.nih.gov/pub/medgen/medgen_pubmed_lnk.txt.gz"
NCBI_GENE_MEDGEN_NAMES = "https://ftp.ncbi.nlm.nih.gov/pub/medgen/NAMES.RRF.gz"

# Protein Atlas #
# https://www.proteinatlas.org/about/download
PROTEIN_ATLAS = "https://www.proteinatlas.org/download/"
PROTEIN_ATLAS_NORMAL_TISSUE = PROTEIN_ATLAS + "normal_tissue.tsv.zip"
PROTEIN_ATLAS_NORMAL_PATHOLOGY = PROTEIN_ATLAS + "pathology.tsv.zip"
PROTEIN_ATLAS_SUBCELLULAR_LOCATIO = PROTEIN_ATLAS + "subcellular_location.tsv.zip"
PROTEIN_ATLAS_RNA_CONSENSUS = PROTEIN_ATLAS + "rna_tissue_consensus.tsv.zip"
PROTEIN_ATLAS_RNA_GTEX_BRAIN = PROTEIN_ATLAS + "rna_brain_gtex.tsv.zip"
PROTEIN_ATLAS_RNA_FANTOM_BRAIN = PROTEIN_ATLAS + "rna_brain_fantom.tsv.zip"
PROTEIN_ATLAS_RNA_SINGLE = PROTEIN_ATLAS + "rna_single_cell_type.tsv.zip"
PROTEIN_ATLAS_RNA_MOUSE_BRAIN_ALLEN = PROTEIN_ATLAS + "rna_mouse_brain_allen.tsv.zip"



"""NCBI API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS

# class TestNcbi:
#     def test_get_gene_by_go(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/gene/by_go?tax_id=9606&gene_id=945&go_id=GO%3A0002765&evidence=IDA&qualifier=involved_in&go_term=immune%20response-inhibiting%20signal%20transduction&category=Process&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_gene_by_mim(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/gene/by_mim?mim_number=159590&gene_id=945&type=gene&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_gene_by_ensembl(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/gene/by_ensembl?tax_id=9606&gene_id=945&ensembl_gene_identifier=ENSG00000105383&rna_nucleotide_accession_version=NM_001082618.2&ensembl_rna_identifier=ENST00000421133.6&protein_accession_version=NP_001076087.1&ensembl_protein_identifier=ENSP00000410126.1&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_gene_info(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/gene/info?gene_id=945&tax_id=9606&symbol=CD33&type_of_gene=protein-coding&chromosome=19&map_location=19q13.41&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_medgen(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/medgen?cui=C4021990&name=Abnormality%20of%20the%20microglia&source=GTR&suppress=N&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_medgen_by_pmid(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/medgen/by_pmid?pmid=29856956&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results
#
#     def test_get_go_by_pmid(self, client):
#         response = client.get(
#             'api/v1/ebel/ncbi/gene/go/by_pmid?pmid=10887109&page_size=10&page=1',
#             content_type='application/json'
#         )
#         output = format_response_data(response)
#         expected_results = {
#             "confidence_value": 0.56,
#             "detection_method": "two hybrid array",
#             "detection_method_psimi_id": 397,
#             "id": 681219,
#             "int_a_uniprot_id": "P20138",
#             "int_b_uniprot_id": "Q9Y5Z9",
#             "interaction_ids": "intact:EBI-23994924|imex:IM-25472-94969",
#             "interaction_type": "physical association",
#             "interaction_type_psimi_id": 915,
#             "pmid": 32296183
#         }
#         assert output[NUM_RESULTS] == 1
#         assert output[PAGE_SIZE] == 10
#         results = output[RESULTS]
#         assert isinstance(results, list)
#         hit = results[0]
#         assert isinstance(hit, dict)
#         assert hit == expected_results



"""ClinVar API methods."""

from math import ceil

from flask import request

from ebel import Bel
from ebel.manager.rdbms.models import clinvar
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_data, _get_paginated_query_result,
                                  _get_pagination,
                                  _get_terms_from_model_starts_with)


def get_clinvar():
    """Get generic ClinVar entry."""
    return _get_data(clinvar.Clinvar)


def get_clinvar_simple():
    """Generic Clinvar result query."""
    cc = clinvar.Clinvar
    cp = clinvar.ClinvarPhenotype
    paras = {
        k: v
        for k, v in request.args.items()
        if k in ["gene_id", "gene_symbol", "hgnc_id", "allele_id", "assembly", "rs_db_snp"]
    }
    query = (
        RDBMS.get_session()
        .query(cc.id, cc.hgnc_id, cc.allele_id, cc.gene_symbol, cc.assembly, cc.rs_db_snp)
        .filter_by(**paras)
        .join(clinvar.clinvar__clinvar_phenotype)
        .join(cp)
    )

    phenotype = request.args.get("phenotype")
    if phenotype:
        query = query.filter_by(phenotype="Alzheimer disease")

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_phenotype_starts_with():
    """Get Clinvar results by fuzzy phenotype search."""
    return _get_terms_from_model_starts_with("phenotype", clinvar.ClinvarPhenotype.phenotype)


def get_by_other_identifier():
    """Get Clinvar results by other identifier."""
    cc = clinvar.Clinvar
    co = clinvar.ClinvarOtherIdentifier

    db = request.args.get("db")
    identifier = request.args.get("identifier")
    query = (
        RDBMS.get_session()
        .query(
            co.db,
            co.identifier,
            cc.id,
            cc.hgnc_id,
            cc.allele_id,
            cc.gene_symbol,
            cc.assembly,
            cc.rs_db_snp,
        )
        .join(co)
        .filter_by(db=db)
        .filter(co.identifier.like(identifier))
    )

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_by_medgen():
    """Get Clinvar results by MedGen identifier."""
    cc = clinvar.Clinvar
    cm = clinvar.ClinvarPhenotypeMedgen

    # db = request.args.get('db')
    identifier = request.args.get("identifier")
    query = (
        RDBMS.get_session()
        .query(
            cm.identifier,
            cc.id,
            cc.hgnc_id,
            cc.allele_id,
            cc.gene_symbol,
            cc.assembly,
            cc.rs_db_snp,
        )
        .join(cm)
        .filter_by(identifier=identifier)
    )

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_ebel_relation():
    """Get ClinVar related e(BE:L) relations."""
    b = Bel()
    p = _get_pagination()
    req = dict(request.args.copy())
    wheres = []
    if req.get("rs_number"):
        wheres.append(f"in.rs_number='{req['rs_number']}'")
    relation_type = req.get("relation_type")
    allowed_relations = [
        "has_snp_clinvar",
        "has_mapped_snp_cv",
        "has_downstream_snp_cv",
        "has_upstream_snp_cv",
    ]
    relation_type = relation_type if relation_type in allowed_relations else allowed_relations[0]
    wheres += [f'out.{k} = "{v}"' for k, v in req.items() if k in ["namespace", "name"]]
    wheres += [f'{k} = "{v}"' for k, v in req.items() if k in ["phenotype", "keyword"]]

    columns = """in.rs_number as rs_number,
                 @class.asString() as relation_type,
                 phenotype,
                 clinical_significance,
                 out.namespace as namespace,
                 out.name as name"""

    from_where_sql = f""" FROM {relation_type}"""
    if wheres:
        from_where_sql += " WHERE " + " AND ".join(wheres)

    number_of_results = b.query_get_dict(f"Select count(*) {from_where_sql}")[0]["count"]

    sql = f"SELECT {columns} {from_where_sql} LIMIT {p.page_size} SKIP {p.skip}"

    return {
        "page": p.page,
        "page_size": p.page_size,
        "number_of_results": number_of_results,
        "pages": ceil(number_of_results / p.page_size),
        "results": b.query_get_dict(sql),
    }



"""BioGRID API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestBiogrid:
    """Unit tests for the Biogrid API module."""

    def test_get_has_ppi_bg_by_symbol_taxid(self, client):
        """Test has_ppi collection."""
        # TODO implement once fixed: currently returns an empty list
        pass

    def test_get_sources(self, client):
        response = client.get("api/v1/biogrid/sources", content_type="application/json")
        output = format_response_data(response)
        expected_results = {"BIOGRID": 1, "FLYBASE": 4, "POMBASE": 2, "WORMBASE": 3}
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_has_ppi_bg_by_uniprot(self, client):
        response = client.get(
            "api/v1/ebel/biogrid/modification/by_uniprot?uniprot=P10636", content_type="application/json"
        )
        output = format_response_data(response)
        example_result = {  # Used to quickly getting cols
            "biogrid_ids": [2623252],
            "modification": "Proteolytic Processing",
            "object_label": "Caspase-7",
            "object_name": "CASP7",
            "object_namespace": "HGNC",
            "object_taxonomy_id": 9606,
            "object_uniprot": "P55210",
            "pmids": [12888622],
            "subject_label": "Microtubule-associated protein tau",
            "subject_name": "MAPT",
            "subject_namespace": "HGNC",
            "subject_taxonomy_id": 9606,
            "subject_uniprot": "P10636",
        }
        assert len(output) > 4
        assert isinstance(output, list)
        hit = None

        for entry in output:
            if entry["object_label"] == "Caspase-7":
                hit = entry
                break

        assert bool(hit)
        assert isinstance(hit, dict)
        for key, val in example_result.items():
            assert hit[key] == val

    def test_get_biogrid_by_biogrid_id(self, client):
        response = client.get("api/v1/biogrid/by_biogrid_id/2658248", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "biogrid_a": {
                "entrez": 59272,
                "symbol": "ACE2",
                "systematic_name": "UNQ868/PRO1885",
                "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
                "trembl": None,
                "uniprot": "Q9BYF1",
            },
            "biogrid_b": {
                "entrez": 43740568,
                "symbol": "S",
                "systematic_name": "GU280_gp02",
                "taxonomy": {
                    "organism_name": "Severe acute respiratory syndrome coronavirus 2",
                    "taxonomy_id": 2697049,
                },
                "trembl": None,
                "uniprot": "P0DTC2",
            },
            "biogrid_id": 2658248,
            "experimental_system": {
                "experimental_system": "Reconstituted Complex",
                "experimental_system_type": "physical",
                "frequency": 58466,
            },
            "modification": None,
            "publication": {
                "author_name": "Sun C ",
                "publication_year": 2020,
                "source": "DOI",
                "source_identifier": "10.1101/2020.02.16.951723",
            },
            "qualification": "The Receptor Binding Domain (RBD) of the SARS-CoV-2 Spike protein interacts with human ACE2 in vitro.",
            "score": None,
            "source": "BIOGRID",
            "throughput": {"frequency": 366952, "throughput": "Low Throughput"},
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_biogrid_by_biogrid_id_using_post(self, client):
        response = client.post(
            "api/v1/biogrid/by_biogrid_id",
            json={"biogrid_id": 2658248},
        )
        output = format_response_data(response)

        expected_results = {
            "biogrid_a": {
                "entrez": 59272,
                "symbol": "ACE2",
                "systematic_name": "UNQ868/PRO1885",
                "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
                "trembl": None,
                "uniprot": "Q9BYF1",
            },
            "biogrid_b": {
                "entrez": 43740568,
                "symbol": "S",
                "systematic_name": "GU280_gp02",
                "taxonomy": {
                    "organism_name": "Severe acute respiratory syndrome coronavirus 2",
                    "taxonomy_id": 2697049,
                },
                "trembl": None,
                "uniprot": "P0DTC2",
            },
            "biogrid_id": 2658248,
            "experimental_system": {
                "experimental_system": "Reconstituted Complex",
                "experimental_system_type": "physical",
                "frequency": 58466,
            },
            "modification": None,
            "publication": {
                "author_name": "Sun C ",
                "publication_year": 2020,
                "source": "DOI",
                "source_identifier": "10.1101/2020.02.16.951723",
            },
            "qualification": "The Receptor Binding Domain (RBD) of the SARS-CoV-2 Spike protein interacts with human ACE2 in vitro.",
            "score": None,
            "source": "BIOGRID",
            "throughput": {"frequency": 366952, "throughput": "Low Throughput"},
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_experimental_systems(self, client):
        response = client.get("api/v1/biogrid/experimental_systems", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "Affinity Capture-Luminescence": 24,
            "Affinity Capture-MS": 2,
            "Affinity Capture-RNA": 9,
            "Affinity Capture-Western": 5,
            "Biochemical Activity": 12,
            "Co-crystal Structure": 21,
            "Co-fractionation": 6,
            "Co-localization": 20,
            "Co-purification": 19,
            "Dosage Growth Defect": 26,
            "Dosage Lethality": 25,
            "Dosage Rescue": 18,
            "FRET": 22,
            "Far Western": 27,
            "Negative Genetic": 1,
            "PCA": 11,
            "Phenotypic Enhancement": 15,
            "Phenotypic Suppression": 14,
            "Positive Genetic": 4,
            "Protein-RNA": 17,
            "Protein-peptide": 23,
            "Proximity Label-MS": 7,
            "Reconstituted Complex": 8,
            "Synthetic Growth Defect": 10,
            "Synthetic Haploinsufficiency": 28,
            "Synthetic Lethality": 13,
            "Synthetic Rescue": 16,
            "Two-hybrid": 3,
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_taxonomies(self, client):
        response = client.get("api/v1/biogrid/taxonomies", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "Anopheles gambiae (PEST)": 180454,
            "Apis mellifera": 7460,
            "Arabidopsis thaliana (Columbia)": 3702,
            "Bacillus subtilis (168)": 224308,
            "Bos taurus": 9913,
            "Caenorhabditis elegans": 6239,
            "Candida albicans (SC5314)": 237561,
            "Canis familiaris": 9615,
            "Cavia porcellus": 10141,
            "Chlamydomonas reinhardtii": 3055,
            "Chlorocebus sabaeus": 60711,
            "Cricetulus griseus": 10029,
            "Danio rerio": 7955,
            "Dictyostelium discoideum (AX4)": 352472,
            "Drosophila melanogaster": 7227,
            "Emericella nidulans (FGSC A4)": 227321,
            "Equus caballus": 9796,
            "Escherichia coli (K12)": 83333,
            "Escherichia coli (K12/MC4100/BW2952)": 595496,
            "Escherichia coli (K12/MG1655)": 511145,
            "Escherichia coli (K12/W3110)": 316407,
            "Felis Catus": 9685,
            "Gallus gallus": 9031,
            "Glycine max": 3847,
            "Hepatitus C Virus": 11103,
            "Homo sapiens": 9606,
            "Human Herpesvirus 1": 10298,
            "Human Herpesvirus 2": 10310,
            "Human Herpesvirus 3": 10335,
            "Human Herpesvirus 4": 10376,
            "Human Herpesvirus 5": 10359,
            "Human Herpesvirus 6A": 32603,
            "Human Herpesvirus 6B": 32604,
            "Human Herpesvirus 7": 10372,
            "Human Herpesvirus 8": 37296,
            "Human Immunodeficiency Virus 1": 11676,
            "Human Immunodeficiency Virus 2": 11709,
            "Human papillomavirus (10)": 333759,
            "Human papillomavirus (16)": 333760,
            "Human papillomavirus (32)": 333763,
            "Human papillomavirus (5)": 333923,
            "Human papillomavirus (6b)": 10600,
            "Human papillomavirus (7)": 10620,
            "Human papillomavirus (9)": 10621,
            "Leishmania major (Friedlin)": 347515,
            "Macaca mulatta": 9544,
            "Meleagris gallopavo": 9103,
            "Middle-East Respiratory Syndrome-related Coronavirus": 1335626,
            "Monodelphis domestica": 13616,
            "Mus musculus": 10090,
            "Mycobacterium tuberculosis (H37Rv)": 83332,
            "Neurospora crassa (OR74A)": 367110,
            "Nicotiana tomentosiformis": 4098,
            "Oryctolagus cuniculus": 9986,
            "Oryza sativa (Japonica)": 39947,
            "Ovis aries": 9940,
            "Pan troglodytes": 9598,
            "Pediculus humanus": 121224,
            "Plasmodium falciparum (3D7)": 36329,
            "Rattus norvegicus": 10116,
            "Ricinus communis": 3988,
            "Saccharomyces cerevisiae (S288c)": 559292,
            "Schizosaccharomyces pombe (972h)": 284812,
            "Selaginella moellendorffii": 88036,
            "Severe acute respiratory syndrome coronavirus 2": 2697049,
            "Severe acute respiratory syndrome-related coronavirus": 694009,
            "Simian Immunodeficiency Virus": 11723,
            "Simian Virus 40": 10633,
            "Solanum lycopersicum": 4081,
            "Solanum tuberosum": 4113,
            "Sorghum bicolor": 4558,
            "Streptococcus pneumoniae (ATCCBAA255)": 171101,
            "Strongylocentrotus purpuratus": 7668,
            "Sus scrofa": 9823,
            "Tobacco Mosaic Virus": 12242,
            "Ustilago maydis (521)": 237631,
            "Vaccinia Virus": 10245,
            "Vitis vinifera": 29760,
            "Xenopus laevis": 8355,
            "Zea mays": 4577,
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_modifications(self, client):
        response = client.get("api/v1/biogrid/modifications", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "Acetylation": 4,
            "Deacetylation": 10,
            "Demethylation": 14,
            "Deneddylation": 15,
            "Dephosphorylation": 9,
            "Desumoylation": 13,
            "Deubiquitination": 5,
            "FAT10ylation": 19,
            "Glycosylation": 17,
            "Methylation": 7,
            "Nedd(Rub1)ylation": 11,
            "Neddylation": 18,
            "No Modification": 3,
            "Phosphorylation": 1,
            "Prenylation": 16,
            "Proteolytic Processing": 6,
            "Ribosylation": 12,
            "Sumoylation": 8,
            "Ubiquitination": 2,
            "de-ISGylation": 20,
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_biogrid_by_pmid(self, client):
        response = client.get("api/v1/biogrid/by_pmid/21685874", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "biogrid_a": {
                "entrez": 853167,
                "symbol": "GCN5",
                "systematic_name": "YGR252W",
                "taxonomy": {"organism_name": "Saccharomyces cerevisiae (S288c)", "taxonomy_id": 559292},
                "trembl": None,
                "uniprot": "Q03330",
            },
            "biogrid_b": {
                "entrez": 852295,
                "symbol": "HHT1",
                "systematic_name": "YBR010W",
                "taxonomy": {"organism_name": "Saccharomyces cerevisiae (S288c)", "taxonomy_id": 559292},
                "trembl": None,
                "uniprot": "P61830",
            },
            "biogrid_id": 543927,
            "experimental_system": {
                "experimental_system": "Biochemical Activity",
                "experimental_system_type": "physical",
                "frequency": 24247,
            },
            "modification": {"frequency": 846, "modification": "Acetylation"},
            "publication": {
                "author_name": "Bian C ",
                "publication_year": 2011,
                "source": "PUBMED",
                "source_identifier": "21685874",
            },
            "qualification": "the purified SAGA complex can acetylate histone 3 in vitro",
            "score": None,
            "source": "BIOGRID",
            "throughput": {"frequency": 1611845, "throughput": "High Throughput"},
        }
        assert isinstance(output, list)
        hit = output[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_biogrid(self, client):
        response = client.get(
            "api/v1/biogrid?id_type_a=symbol&interactor_a=MAPT&taxonomy_id_a=9606&id_type_b=symbol&taxonomy_id_b=9606&modification=Acetylation&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "biogrid_a": {
                "entrez": 2033,
                "symbol": "EP300",
                "systematic_name": "RP1-85F18.1",
                "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
                "trembl": "Q7Z6C1",
                "uniprot": "Q09472",
            },
            "biogrid_b": {
                "entrez": 4137,
                "symbol": "MAPT",
                "systematic_name": None,
                "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
                "trembl": "B3KTM0",
                "uniprot": "P10636",
            },
            "biogrid_id": 558818,
            "experimental_system": {
                "experimental_system": "Biochemical Activity",
                "experimental_system_type": "physical",
                "frequency": 24247,
            },
            "modification": {"frequency": 846, "modification": "Acetylation"},
            "publication": {
                "author_name": "Min SW ",
                "publication_year": 2010,
                "source": "PUBMED",
                "source_identifier": "20869593",
            },
            "qualification": None,
            "score": None,
            "source": "BIOGRID",
            "throughput": {"frequency": 366952, "throughput": "Low Throughput"},
        }
        assert output[NUM_RESULTS] == 4
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_biogrid_using_post(self, client):
        response = client.post(
            "api/v1/biogrid",
            json={"experimental_system_id": 2, "page": 2, "page_size": 3},
            follow_redirects=True,
        )
        output = format_response_data(response)

        expected_results = {
            "biogrid_a": {
                "entrez": 1489671,
                "symbol": "E",
                "systematic_name": "SARS-CoV-4",
                "taxonomy": {
                    "organism_name": "Severe acute respiratory syndrome-related coronavirus",
                    "taxonomy_id": 694009,
                },
                "trembl": None,
                "uniprot": "P59637",
            },
            "biogrid_b": {
                "entrez": 83871,
                "symbol": "RAB34",
                "systematic_name": None,
                "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
                "trembl": "B4DNC0",
                "uniprot": "Q9BZG1",
            },
            "biogrid_id": 2754699,
            "experimental_system": {
                "experimental_system": "Affinity Capture-MS",
                "experimental_system_type": "physical",
                "frequency": 459921,
            },
            "modification": None,
            "publication": {
                "author_name": "Stukalov A ",
                "publication_year": 2020,
                "source": "DOI",
                "source_identifier": "10.1101/2020.06.17.156455",
            },
            "qualification": "Affinity capture-MS was carried out using HA-tagged viral proteins as baits in human A549 lung carcinoma cells. Significant interactions were identified as those where the prey protein was at least 4 times enriched against the background (median log2 score > 2) with a p-value <= 1E-3.",
            "score": 8.15622,
            "source": "BIOGRID",
            "throughput": {"frequency": 1611845, "throughput": "High Throughput"},
        }
        assert output[NUM_RESULTS] > 459000
        assert output[PAGE_SIZE] == 3
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_interactor_by_symbol_starts_with(self, client):
        response = client.get(
            "api/v1/biogrid/interactor/by_symbol_starts_with?symbol=trem&taxonomy_id=9606&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "entrez": 54210,
            "symbol": "TREM1",
            "systematic_name": None,
            "taxonomy": {"organism_name": "Homo sapiens", "taxonomy_id": 9606},
            "trembl": "Q38L15",
            "uniprot": "Q9NP99",
        }
        assert output[NUM_RESULTS] == 4
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""HGNC API methods."""
from flask import request

from ebel.manager.rdbms.models import hgnc
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import _get_data


def get_by_symbol():
    """Get paginated list of HGNC entries by given gene symbol."""
    symbol = request.args.get("symbol")
    if symbol:
        db_entry = RDBMS.get_session().query(hgnc.Hgnc).filter_by(symbol=symbol).first()
        if db_entry:
            return db_entry.as_dict()


def get_uniprot_accession_by_hgnc_symbol():
    """Return UniProt accession number by HGCN gene symbol."""
    symbol = request.args.get("symbol")
    if symbol:
        db_entry = RDBMS.get_session().query(hgnc.Hgnc).filter_by(symbol=symbol).first()
        if db_entry:
            return db_entry.uniprots[0].accession


def get_hgnc():
    """Get paginated list of HGNC entries."""
    return _get_data(hgnc.Hgnc)


def get_gene_group_name():
    """Get paginated list of HGNC entries."""
    return _get_data(hgnc.GeneGroupName)



"""Collection of methods for handling information caching."""
import getpass
import logging

import pymysql

from ebel import defaults
from ebel.config import write_to_config

# TODO: Decide whether to use these methods or those from database

logger = logging.getLogger(__name__)


def set_mysql_interactive() -> tuple:
    """Interactive mode to setup MySQL database."""
    print(
        "Interactive mode\n \
          ================\n \
          1st setup of db and user with root:\n"
    )
    host = input("host[localhost]:") or "localhost"
    user = input("ebel user[ebel]:") or "ebel"
    password = getpass.getpass(prompt="ebel password[ebel]:") or "ebel"
    db = input("database name[ebel]") or "ebel"
    print(
        "If you want to setup the database automatically,\n \
          then type in the root password, otherwise nothing"
    )
    root_pwd = getpass.getpass(prompt="root password (only for 1st setup):")

    if root_pwd:
        root_host = getpass.getpass(prompt="IP or name mysql server [localhost]:") or "localhost"
        conn = pymysql.connect(host=root_host, user="root", password=root_pwd)
        c = conn.cursor()
        db_exists = c.execute("show databases like '{}'".format(db))

        if not db_exists:
            c.execute("CREATE DATABASE {} CHARACTER SET utf8".format(db))

        else:
            logger.warning(f"Database '{db}' already exists!")

        user_exists = c.execute("Select 1 from mysql.user where User='{}'".format(user))

        if not user_exists:
            sql = "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".format(
                user,
                host,
                password,
            )
            c.execute(sql)
        else:
            logger.warning(f"Database '{db}' already exists!")

        privileges_exists = c.execute("Select 1 from mysql.db where User='{}' and Db='{}'".format(user, db))
        if not privileges_exists:
            c.execute("GRANT ALL PRIVILEGES ON {}.* TO '{}'@'%'  IDENTIFIED BY '{}'".format(db, user, password))
        else:
            logger.warning(f"User already has privileges for database '{db}'")

        c.execute("FLUSH PRIVILEGES")

    return host, user, password, db


def set_mysql_connection(
    host: str = "localhost",
    user: str = "ebel_user",
    password: str = "ebel_passwd",
    db: str = "ebel",
    charset: str = "utf8mb4",
):
    """Set the connection using MySQL Parameters.

    Parameters
    ----------
    host : str
        MySQL database host.
    user : str
        MySQL database user.
    password : str
        MySQL database password.
    db : str
        MySQL database name.
    charset : str
        MySQL database charset.

    Returns
    -------
    str
        SQLAlchemy MySQL connection string.

    """
    connection_string = "mysql+pymysql://{user}:{passwd}@{host}/{db}?charset={charset}".format(
        host=host, user=user, passwd=password, db=db, charset=charset
    )
    set_connection(connection_string)

    return connection_string


def set_always_create_new_db(always_create_new_db: bool = True) -> None:
    """Set in configuration option `always_create_new_db`.

    Parameters
    ----------
    always_create_new_db : bool
        Option `always_create_new_db` in section `database` in config file.

    """
    write_to_config("database", "always_create_new", str(always_create_new_db))


def set_connection(connection: str = defaults.CONN_STR_DEFAULT) -> None:
    """Set the connection string for SQLAlchemy.

    Parameters
    ----------
    connection: str
        SQLAlchemy connection string.

    """
    write_to_config("database", "sqlalchemy_connection_string", connection)



"""HGNC API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestHgnc:
    def test_get_by_symbol(self, client):
        response = client.get("api/v1/hgnc/by_symbol?symbol=CD33", content_type="application/json")
        output = format_response_data(response)
        expected_results = {
            "agr": "HGNC:1659",
            "alias_names": ["sialic acid binding Ig-like lectin 3"],
            "alias_symbols": ["SIGLEC3", "SIGLEC-3", "p67", "FLJ00391"],
            "bioparadigms_slc": None,
            "ccdss": ["CCDS46157", "CCDS54299", "CCDS33084"],
            "cd": "CD33",
            "cosmic": None,
            "date_approved_reserved": "1986-01-01",
            "date_modified": "2016-10-05",
            "date_name_changed": "2006-03-28",
            "date_symbol_changed": None,
            "enas": ["M23197"],
            "ensembl_gene_id": "ENSG00000105383",
            "entrez_id": 945,
            "enzymes": [],
            "gene_group_ids": [471, 590, 745],
            "gene_group_names": [
                "CD molecules",
                "V-set domain containing",
                "Sialic acid binding Ig like lectins",
            ],
            "hgnc_id": "HGNC:1659",
            "homeodb": None,
            "horde_id": None,
            "imgt": None,
            "iuphar": "objectId:2601",
            "kznf_gene_catalog": None,
            "lncipedia": None,
            "lncrnadb": None,
            "location": "19q13.41",
            "location_sortable": "19q13.41",
            "locus_group": "protein-coding gene",
            "locus_type": "gene with protein product",
            "lsdbs": [],
            "merops": None,
            "mgds": ["MGI:99440"],
            "mirbase": None,
            "name": "CD33 molecule",
            "omims": [159590],
            "orphanet": None,
            "pre_symbols": [],
            "prev_names": ["CD33 antigen (gp67)"],
            "pubmeds": [3139766, 9465907],
            "refseqs": ["NM_001772"],
            "rgds": ["RGD:1596020"],
            "rna_centrals": [],
            "snornabase": None,
            "status": "Approved",
            "symbol": "CD33",
            "ucsc_id": "uc002pwa.3",
            "uniprots": ["P20138"],
            "uuid": "982e665a-2cc7-4ef3-b9f6-4406e08ed6c8",
            "vega_id": "OTTHUMG00000182891",
            "version": 1702607059790331904,
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_uniprot_accession_by_hgnc_symbol(self, client):
        response = client.get(
            "api/v1/hgnc/uniprot_accession_by_hgnc_symbol?symbol=CD33",
            content_type="application/json",
        )
        output = format_response_data(response)
        assert isinstance(output, str)
        assert output == "P20138"

    def test_get_hgnc(self, client):
        response = client.get(
            "api/v1/hgnc?symbol=CD33&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "agr": "HGNC:1659",
            "alias_names": ["sialic acid binding Ig-like lectin 3"],
            "alias_symbols": ["SIGLEC3", "SIGLEC-3", "p67", "FLJ00391"],
            "bioparadigms_slc": None,
            "ccdss": ["CCDS46157", "CCDS54299", "CCDS33084"],
            "cd": "CD33",
            "cosmic": None,
            "date_approved_reserved": "1986-01-01",
            "date_modified": "2016-10-05",
            "date_name_changed": "2006-03-28",
            "date_symbol_changed": None,
            "enas": ["M23197"],
            "ensembl_gene_id": "ENSG00000105383",
            "entrez_id": 945,
            "enzymes": [],
            "gene_group_ids": [471, 590, 745],
            "gene_group_names": [
                "CD molecules",
                "V-set domain containing",
                "Sialic acid binding Ig like lectins",
            ],
            "hgnc_id": "HGNC:1659",
            "homeodb": None,
            "horde_id": None,
            "imgt": None,
            "iuphar": "objectId:2601",
            "kznf_gene_catalog": None,
            "lncipedia": None,
            "lncrnadb": None,
            "location": "19q13.41",
            "location_sortable": "19q13.41",
            "locus_group": "protein-coding gene",
            "locus_type": "gene with protein product",
            "lsdbs": [],
            "merops": None,
            "mgds": ["MGI:99440"],
            "mirbase": None,
            "name": "CD33 molecule",
            "omims": [159590],
            "orphanet": None,
            "pre_symbols": [],
            "prev_names": ["CD33 antigen (gp67)"],
            "pubmeds": [3139766, 9465907],
            "refseqs": ["NM_001772"],
            "rgds": ["RGD:1596020"],
            "rna_centrals": [],
            "snornabase": None,
            "status": "Approved",
            "symbol": "CD33",
            "ucsc_id": "uc002pwa.3",
            "uniprots": ["P20138"],
            "uuid": "982e665a-2cc7-4ef3-b9f6-4406e08ed6c8",
            "vega_id": "OTTHUMG00000182891",
            "version": 1702607059790331904,
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""UniProt API methods."""
import json

from flask import request

from ebel import Bel
from ebel.manager.rdbms.models import uniprot
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_paginated_query_result,
                                  _get_terms_from_model_starts_with)

from . import add_query_filters

model_by_tablename = {x.__tablename__: x for x in uniprot.Base.__subclasses__()}
nm_tables = {
    "uniprot_keyword": uniprot.uniprot__uniprot_keyword,
    "uniprot_host": uniprot.uniprot__uniprot_host,
    "uniprot_xref": uniprot.uniprot__uniprot_xref,
    "uniprot_subcellular_location": uniprot.uniprot__uniprot_subcellular_location,
}


def get_uniprot_advanced():
    """Get uniprot results using advanced joins."""
    data = json.loads(request.data)
    up = uniprot.Uniprot
    return_columns = (
        up.accession,
        up.name,
        uniprot.Function.description,
        up.recommended_name,
        uniprot.GeneSymbol.symbol,
        uniprot.Organism.scientific_name,
    )
    query = (
        RDBMS.get_session()
        .query(up)
        .outerjoin(uniprot.Function)
        .outerjoin(uniprot.GeneSymbol)
        .outerjoin(uniprot.Organism)
    )
    already_joined_models = (up, uniprot.Function, uniprot.GeneSymbol, uniprot.Organism)

    for table, columns in data.items():
        if table not in ("page", "page_size", "number_of_results"):
            model = model_by_tablename.get(table)
            values_exists = any([x.get("value") for x in columns.values()])
            if model and model not in already_joined_models and values_exists:
                if table in nm_tables:
                    query = query.outerjoin(nm_tables[table])

                query = query.outerjoin(model)
            if values_exists:
                query = add_query_filters(query, columns, model)

    query = query.with_entities(*return_columns)
    query = query.group_by(up.id)
    return _get_paginated_query_result(query, return_dict=True, print_sql=True)


def get_uniprot():
    """Get paginated list of UniProt entries."""
    q = RDBMS.get_session().query(uniprot.Uniprot)

    accession = request.args.get("accession")
    q = q.filter_by(accession=accession) if accession else q

    taxid = request.args.get("taxonomy_id")
    q = q.filter_by(taxid=taxid) if taxid else q

    gene_symbol = request.args.get("gene_symbol")
    if gene_symbol:
        q = q.join(uniprot.GeneSymbol).filter_by(symbol=gene_symbol)

    keyword = request.args.get("keyword")
    if keyword:
        q = q.join(uniprot.uniprot__uniprot_keyword).join(uniprot.Keyword).filter_by(keyword_name=keyword)

    xref_db = request.args.get("xref_db")
    xref_id = request.args.get("xref_id")
    if xref_db or xref_id:
        q = q.join(uniprot.uniprot__uniprot_xref).join(uniprot.Xref)
        q = q.filter_by(db=xref_db) if xref_db else q
        q = q.filter_by(identifier=xref_id) if xref_id else q

    subcellular_location = request.args.get("subcellular_location")
    if subcellular_location:
        q = (
            q.join(uniprot.uniprot__uniprot_subcellular_location)
            .join(uniprot.SubcellularLocation)
            .filter_by(name=subcellular_location)
        )

    return _get_paginated_query_result(q)


def get_keyword_starts_with():
    """Get entries where keyword starts with given value."""
    return _get_terms_from_model_starts_with("keyword", uniprot.Keyword.keyword_name)


def get_subcellular_location_starts_with():
    """Get entries where subcellular location starts with given value."""
    return _get_terms_from_model_starts_with("subcellular_location", uniprot.SubcellularLocation.name)


def get_gene_symbol_starts_with():
    """Get entries where symbol starts with given value."""
    return _get_terms_from_model_starts_with("gene_symbol", uniprot.GeneSymbol.symbol)


def get_gene_starts_with():
    """Get entries where gene starts with given value."""
    return _get_terms_from_model_starts_with("gene", uniprot.Gene.name)


def get_organism_starts_with():
    """Get entries where organism starts with given value."""
    return _get_terms_from_model_starts_with("organism", uniprot.Organism.scientific_name)


def get_function_starts_with():
    """Get entries where description starts with given value."""
    return _get_terms_from_model_starts_with("description", uniprot.Function.description)


def get_bel_node_uniprot():
    """Get UniProt related eBEL nodes."""
    b = Bel()
    conf = {
        "rid": "@rid.asString()",
        "name": "name",
        "namespace": "namespace",
        "bel": "bel",
        "uniprot_accession": "uniprot",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM protein WHERE pure = true"

    uniprot = request.args.get("accession")

    if uniprot:
        sql += f' AND uniprot = "{uniprot}"'
    else:
        return {"error": "uniprot_accession is required."}

    entries = b.query_get_dict(sql)
    if entries:
        result_dict = entries[0]
        result_dict["edges"] = {}
        for direction in ("in", "out"):
            match = "match {class:protein, where:(uniprot='"
            match += uniprot + "' and pure=true)}." + direction + "E(bel_relation){as:e}"
            match += " return e.@rid, e.@class as relation"
            sql_edges = f"""Select relation, count(*)
                    from ({match}) group by relation order by count desc"""
            result_dict["edges"][direction] = b.query_get_dict(sql_edges)

        return result_dict



"""NSIDES API unit tests."""

from .conftest import format_response_data
from .constants import NUM_RESULTS, PAGE_SIZE, RESULTS


class TestNsides:
    def test_get_nsides(self, client):
        response = client.get(
            "api/v1/nsides?drug_rxnorn_id=4024&drug_concept_name=ergoloid%20mesylates%2C%20USP&condition_meddra_id=10002034&condition_concept_name=Anaemia&a=6&b=126&c=21&d=1299&prr=2.85714&prr_error=0.45382&mean_reporting_frequency=0.0454545&page_size=10&page=1",
            content_type="application/json",
        )
        output = format_response_data(response)
        expected_results = {
            "a": 6,
            "b": 126,
            "c": 21,
            "condition_concept_name": "Anaemia",
            "condition_meddra_id": 10002034,
            "d": 1299,
            "drug_concept_name": "ergoloid mesylates, USP",
            "drug_rxnorn_id": "4024",
            "id": 1,
            "mean_reporting_frequency": 0.0454545,
            "prr": 2.85714,
            "prr_error": 0.45382,
            "source": "offsides",
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""HGNC RDBMS model definition."""
from sqlalchemy import (BigInteger, Column, Date, ForeignKey, Integer, String,
                        Text)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

from ebel.manager.rdbms.models import object_as_dict

Base = declarative_base()


class Hgnc(Base):
    """Class definition for the hgnc table."""

    __tablename__ = "hgnc"
    id = Column(Integer, primary_key=True)
    hgnc_id = Column(String(20))
    version = Column(BigInteger)
    bioparadigms_slc = Column(String(20))
    cd = Column(String(20))
    cosmic = Column(String(50))
    date_approved_reserved = Column(Date)
    date_modified = Column(Date)
    date_name_changed = Column(Date)
    date_symbol_changed = Column(Date)
    ensembl_gene_id = Column(String(20))
    entrez_id = Column(Integer)
    homeodb = Column(Integer)
    horde_id = Column(String(50))
    imgt = Column(String(50))
    iuphar = Column(String(50))
    kznf_gene_catalog = Column(Integer)
    lncipedia = Column(String(50))
    lncrnadb = Column(String(50))
    location = Column(String(100))
    location_sortable = Column(String(100))
    locus_group = Column(String(50))
    locus_type = Column(String(50))
    merops = Column(String(20))
    mirbase = Column(String(20))
    name = Column(String(255))
    orphanet = Column(Integer)
    snornabase = Column(String(20))
    status = Column(String(50))
    symbol = Column(String(100), index=True)
    ucsc_id = Column(String(50))
    uuid = Column(String(50))
    vega_id = Column(String(50))
    agr = Column(String(50))
    kznf_gene_catalog = Column(Text)

    pre_symbols = relationship("PrevSymbol", back_populates="hgnc")
    alias_names = relationship("AliasName", back_populates="hgnc")
    alias_symbols = relationship("AliasSymbol", back_populates="hgnc")
    ccdss = relationship("Ccds", back_populates="hgnc")
    enas = relationship("Ena", back_populates="hgnc")
    enzymes = relationship("Enzyme", back_populates="hgnc")
    gene_group_names = relationship("GeneGroupName", back_populates="hgnc")
    gene_group_ids = relationship("GeneGroupId", back_populates="hgnc")
    uniprots = relationship("UniProt", back_populates="hgnc")
    rna_centrals = relationship("RnaCentral", back_populates="hgnc")
    rgds = relationship("Rgd", back_populates="hgnc")
    refseqs = relationship("RefSeq", back_populates="hgnc")
    pubmeds = relationship("PubMed", back_populates="hgnc")
    prev_names = relationship("PrevName", back_populates="hgnc")
    omims = relationship("Omim", back_populates="hgnc")
    mgds = relationship("Mgd", back_populates="hgnc")
    lsdbs = relationship("Lsdb", back_populates="hgnc")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "hgnc_id": self.hgnc_id,
            "version": self.version,
            "bioparadigms_slc": self.bioparadigms_slc,
            "cd": self.cd,
            "cosmic": self.cosmic,
            "date_approved_reserved": self.date_approved_reserved,
            "date_modified": self.date_modified,
            "date_name_changed": self.date_name_changed,
            "date_symbol_changed": self.date_symbol_changed,
            "ensembl_gene_id": self.ensembl_gene_id,
            "entrez_id": self.entrez_id,
            "homeodb": self.homeodb,
            "horde_id": self.horde_id,
            "imgt": self.imgt,
            "iuphar": self.iuphar,
            "kznf_gene_catalog": self.kznf_gene_catalog,
            "lncipedia": self.lncipedia,
            "lncrnadb": self.lncrnadb,
            "location": self.location,
            "location_sortable": self.location_sortable,
            "locus_group": self.locus_group,
            "locus_type": self.locus_type,
            "merops": self.merops,
            "mirbase": self.mirbase,
            "name": self.name,
            "orphanet": self.orphanet,
            "snornabase": self.snornabase,
            "status": self.status,
            "symbol": self.symbol,
            "ucsc_id": self.ucsc_id,
            "uuid": self.uuid,
            "vega_id": self.vega_id,
            "agr": self.agr,
            "pre_symbols": [x.prev_symbol for x in self.pre_symbols],
            "alias_names": [x.alias_name for x in self.alias_names],
            "alias_symbols": [x.alias_symbol for x in self.alias_symbols],
            "ccdss": [x.identifier for x in self.ccdss],
            "enas": [x.identifier for x in self.enas],
            "enzymes": [x.ec_number for x in self.enzymes],
            "gene_group_names": [x.name for x in self.gene_group_names],
            "gene_group_ids": [x.identifier for x in self.gene_group_ids],
            "uniprots": [x.accession for x in self.uniprots],
            "rna_centrals": [x.identifier for x in self.rna_centrals],
            "rgds": [x.identifier for x in self.rgds],
            "refseqs": [x.accession for x in self.refseqs],
            "pubmeds": [x.pmid for x in self.pubmeds],
            "prev_names": [x.prev_name for x in self.prev_names],
            "omims": [x.identifier for x in self.omims],
            "mgds": [x.identifier for x in self.mgds],
            "lsdbs": [x.identifier for x in self.lsdbs],
        }


class PrevSymbol(Base):
    """Class definition for the hgnc_prev_symbol table."""

    __tablename__ = "hgnc_prev_symbol"
    id = Column(Integer, primary_key=True)

    prev_symbol = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="pre_symbols")

    def __str__(self):
        return self.prev_symbol


class AliasName(Base):
    """Class definition for the hgnc_alias_name table."""

    __tablename__ = "hgnc_alias_name"
    id = Column(Integer, primary_key=True)

    alias_name = Column(String(255))

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="alias_names")

    def __str__(self):
        return self.alias_name


class AliasSymbol(Base):
    """Class definition for the hgnc_alias_symbol table."""

    __tablename__ = "hgnc_alias_symbol"
    id = Column(Integer, primary_key=True)

    alias_symbol = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="alias_symbols")

    def __str__(self):
        return self.alias_symbol


class Ccds(Base):
    """Class definition for the hgnc_ccds table."""

    __tablename__ = "hgnc_ccds"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="ccdss")

    def __str__(self):
        return self.identifier


class Ena(Base):
    """Class definition for the hgnc_ena table."""

    __tablename__ = "hgnc_ena"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="enas")

    def __str__(self):
        return self.identifier


class Enzyme(Base):
    """Class definition for the hgnc_enzyme table."""

    __tablename__ = "hgnc_enzyme"
    id = Column(Integer, primary_key=True)

    ec_number = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="enzymes")

    def __str__(self):
        return self.ec_number


class GeneGroupName(Base):
    """Class definition for the hgnc_gene_group_name table."""

    __tablename__ = "hgnc_gene_group_name"
    id = Column(Integer, primary_key=True)

    name = Column(String(255))

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="gene_group_names")

    def __str__(self):
        return self.name

    def as_dict(self):
        """Return obj as dict."""
        return object_as_dict(self)


class GeneGroupId(Base):
    """Class definition for the hgnc_gene_group_id table."""

    __tablename__ = "hgnc_gene_group_id"
    id = Column(Integer, primary_key=True)

    identifier = Column(Integer)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="gene_group_ids")

    def __str__(self):
        return self.identifier


class UniProt(Base):
    """Class definition for the hgnc_uniprot table."""

    __tablename__ = "hgnc_uniprot"
    id = Column(Integer, primary_key=True)

    accession = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="uniprots")

    def __str__(self):
        return self.accession


class RnaCentral(Base):
    """Class definition for the hgnc_rna_central table."""

    __tablename__ = "hgnc_rna_central"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="rna_centrals")

    def __str__(self):
        return self.identifier


class Rgd(Base):
    """Class definition for the hgnc_rgd table."""

    __tablename__ = "hgnc_rgd"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="rgds")

    def __str__(self):
        return self.identifier


class RefSeq(Base):
    """Class definition for the hgnc_refseq table."""

    __tablename__ = "hgnc_refseq"
    id = Column(Integer, primary_key=True)

    accession = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="refseqs")

    def __str__(self):
        return self.accession


class PubMed(Base):
    """Class definition for the hgnc_pubmed table."""

    __tablename__ = "hgnc_pubmed"
    id = Column(Integer, primary_key=True)

    pmid = Column(Integer, index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="pubmeds")

    def __str__(self):
        return self.pmid


class PrevName(Base):
    """Class definition for the hgnc_prev_name table."""

    __tablename__ = "hgnc_prev_name"
    id = Column(Integer, primary_key=True)

    prev_name = Column(String(255))

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="prev_names")

    def __str__(self):
        return self.prev_name


class Omim(Base):
    """Class definition for the hgnc_omim table."""

    __tablename__ = "hgnc_omim"
    id = Column(Integer, primary_key=True)

    identifier = Column(Integer, index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="omims")

    def __str__(self):
        return self.identifier


class Mgd(Base):
    """Class definition for the hgnc_mgd table."""

    __tablename__ = "hgnc_mgd"
    id = Column(Integer, primary_key=True)

    identifier = Column(String(50), index=True)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="mgds")

    def __str__(self):
        return self.identifier


class Lsdb(Base):
    """Class definition for the hgnc_lsdb table."""

    __tablename__ = "hgnc_lsdb"
    id = Column(Integer, primary_key=True)

    identifier = Column(Text)

    hgnc_id = Column(Integer, ForeignKey("hgnc.id"))
    hgnc = relationship("Hgnc", back_populates="lsdbs")

    def __str__(self):
        return self.identifier



"""Generic e(BE:L) API methods."""
import json
from collections import Counter

import numpy as np
import pandas as pd
from flask import request

from ebel import Bel
from ebel.manager.rdbms.models.uniprot import Uniprot
from ebel.web.api import RDBMS


def get_bel_edge_statistics_by_uniprot_accession():
    """Get edge statatistics by UniProt accession number. 'has__' edges are excluded."""
    query_object = json.loads(request.data)
    acc = query_object["uniprot_accession"]
    eclass = query_object["edge_class"]
    results = {}
    for direction in ("out", "in"):
        results[direction] = __get_bel_edge_statistics_by_uniprot_accession(acc, eclass, direction)

    return results


def __get_bel_edge_statistics_by_uniprot_accession(acc, eclass, direction):
    sql_direction = {"out": ("out", "in"), "in": ("in", "out")}

    sql = "match {class:protein, where:(uniprot='%s')}.%sE()" % (
        acc,
        sql_direction[direction][0],
    )
    sql += "{class:%s, as:e,where:(not @class like 'has__%%')}.%sV(){as:o}" % (
        eclass,
        sql_direction[direction][1],
    )
    sql += (
        " return e.@class.asString() as eclass, e.@rid.asString() as erid, o.@class as oclass, "
        "o.@rid.asString() as orid"
    )

    rows = [x.oRecordData for x in Bel().client.command(sql)]
    edge_counter = Counter([x["eclass"] for x in rows])
    edge_counter_sorted = sorted(edge_counter.items(), key=lambda item: item[1], reverse=True)
    edge_counter_ordered = [{"name": x[0], "value": x[1]} for x in edge_counter_sorted]
    # make object unique
    object_rid_class_dict = {x["orid"]: x["oclass"] for x in rows}
    counter = Counter(object_rid_class_dict.values())
    object_counter_ordered = [
        {"name": x[0], "value": x[1]} for x in sorted(counter.items(), key=lambda i: i[1], reverse=True)
    ]
    return {"edges": edge_counter_ordered, "objects": object_counter_ordered}


def _describe(s: pd.Series):
    return {k: (int(v) if isinstance(v, np.int64) else v) for k, v in s.describe().to_dict().items()}


def get_intact_by_uniprot():
    """Get IntAct entries by UniProt ID."""
    query_object = json.loads(request.data)
    columns = [
        "name",
        "recommended_name",
        "accession",
        "detection_method",
        "interaction_type",
        "confidence_value",
        "pmid",
    ]
    acc = query_object["uniprot_accession"]
    sql = f"""(Select u.name, u.recommended_name, i.int_b_uniprot_id, i.detection_method,
     i.interaction_type, i.confidence_value, CAST(i.pmid AS CHAR) from intact i inner join uniprot u on
     (u.accession=i.int_b_uniprot_id) where
     i.int_a_uniprot_id = '{acc}') union
     (Select u.name,u.recommended_name, i.int_a_uniprot_id as uniprot_interaction_partner, i.detection_method,
     i.interaction_type, i.confidence_value, CAST(i.pmid AS CHAR) from intact i inner join uniprot u on
     (u.accession=i.int_a_uniprot_id) where i.int_b_uniprot_id = '{acc}') order by confidence_value desc"""
    df = pd.DataFrame(RDBMS.get_session().execute(sql), columns=columns)
    # rows = [dict(zip(columns, row)) for row in ]
    # statistics_name_10 = Counter([x['name'] for x in rows]).most_common(10)

    r = {
        "rows": df.to_dict("records"),
        "total": df.shape[0],
        "statistics": {
            "confidence_value": _describe(df.confidence_value),
            "name": _describe(df.name),
            "pmid": _describe(df.pmid),
            "detection_method": _describe(df.detection_method),
            "interaction_type": _describe(df.interaction_type),
        },
    }
    return r


def find_all():
    """Get all UniProt entries."""
    term = json.loads(request.data)["term"].strip()
    uniprot_entries = (
        RDBMS.get_session()
        .query(Uniprot.recommended_name)
        .filter(Uniprot.recommended_name.like("%" + term + "%"))
        .count()
    )
    return uniprot_entries



"""KEGG implemention. Dependent on HGNC."""

import os
import re
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict

import pandas as pd
import requests
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.config import get_config_value
from ebel.defaults import DATA_DIR
from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.hgnc import Hgnc
from ebel.manager.orientdb.constants import KEGG
from ebel.manager.orientdb.odb_defaults import BelPmod
from ebel.manager.rdbms.models import kegg


class Kegg(odb_meta.Graph):
    """KEGG class.

    Parses KEGG Markup Language (KGML, https://www.genome.jp/kegg/xml/docs/) and translates to ebel edges.
    """

    def __init__(self, client: OrientDB = None):
        """Init KEGG."""
        self.client = client
        self.biodb_name = KEGG
        self.urls = {"kegg_path_list": urls.KEGG_PATH_LIST}

        self.file_paths = {"kgml": Path(DATA_DIR, self.biodb_name, "kgml")}
        self.file_paths["kgml"].mkdir(exist_ok=True, parents=True)

        super().__init__(
            tables_base=kegg.Base,
            edges=odb_structure.kegg_edges,
            urls=self.urls,
            biodb_name=self.biodb_name,
        )

        species_list = get_config_value(section="KEGG", option="species", value="hsa, dme, mmu, rno")
        self.species = [species_id.strip() for species_id in species_list.split(",")]

        # TODO Make so it builds dict from info from KEGG directly. JSON url in urls.py, needs method to parse
        self.species_dict = {"hsa": 9606, "dme": 7227, "mmu": 10090, "rno": 10116}
        self.tax_namespace_dict = {
            9606: "HGNC",
            10090: "MGI",
            10116: "RGD",
            7227: "DMEL",
        }
        self.species_namespace_dict = {k: self.tax_namespace_dict[v] for k, v in self.species_dict.items()}
        self.hgnc = Hgnc(self.client)
        self.bel_rid_cache = {}

    def __len__(self) -> int:
        """Get number of edges in OrientDB."""
        return self.number_of_edges

    def __contains__(self, rs_number: int) -> bool:
        # TODO: To be implemented
        return True

    @staticmethod
    def get_entries(kegg_xml) -> dict:
        """Return all KEGG XML entries."""
        root = ET.fromstring(kegg_xml)
        dict_entry = {}
        for entry in root.iter("entry"):
            attribute = entry.attrib
            if "link" in attribute:
                dict_entry[attribute["id"]] = [
                    attribute["name"].split(" "),
                    attribute["type"],
                    attribute["link"],
                ]
            else:
                dict_entry[attribute["id"]] = [
                    attribute["name"].split(" "),
                    attribute["type"],
                    "NA",
                ]

        return dict_entry

    @staticmethod
    def get_relations(kegg_xml: str) -> list:
        """Return a list of dictionaries containing information on every KEGG relation.

        Parameters
        ----------
        kegg_xml : str
            A parsed kgml file represented as a string

        Returns
        -------
        relations : list
            List of dictionaries. Each dictionary has information on one KEGG relation in the kgml file
        """
        root = ET.fromstring(kegg_xml)
        relations = []
        for rel in root.iter("relation"):
            relation_dict = dict()
            relation_dict["entry1"] = rel.attrib["entry1"]
            relation_dict["entry2"] = rel.attrib["entry2"]
            relation_dict["kegg_int_type"] = rel.attrib["type"]
            relation_dict["interaction_type"] = [name.attrib["name"] for name in rel.iter("subtype")]
            relations.append(relation_dict)

        return relations

    @staticmethod
    def get_kegg_gene_identifiers(kegg_species_id: str) -> dict:
        """Return a dict of KEGG genes and their IDs."""
        url_kegg_genes = f"http://rest.kegg.jp/list/{kegg_species_id}"
        df_kegg_genes = pd.read_csv(
            url_kegg_genes,
            sep="\t",
            names=["kegg_id", "external_id"],
        )
        dict_kegg_genes = df_kegg_genes.groupby("kegg_id")["external_id"].apply(list).to_dict()

        # Split by ',' or ';' and take the first gene name
        for kegg_id, values in dict_kegg_genes.items():
            dict_kegg_genes[kegg_id] = re.split(r"[,;]+", values[0])[0]

        return dict_kegg_genes

    def _get_pathway_kgml(self, pathway: str):
        """Reads KGML pathway file if downloaded, downloads if not there. Returns XML content."""
        kgml_path = self.file_paths["kgml"].joinpath(pathway)

        if kgml_path.is_file():
            with open(kgml_path, "r") as kgmlf:
                xml_content = kgmlf.read()

        else:
            url_path = f"http://rest.kegg.jp/get/{pathway}/kgml"
            xml_content = requests.get(url_path).text
            with open(kgml_path, "w") as kgmlf:
                kgmlf.write(xml_content)

        return xml_content

    def insert_data(self) -> Dict[str, int]:
        """Insert KEGG data into database."""
        dfs = []

        for kegg_species_identifier in tqdm(self.species, desc=f"Import {self.biodb_name.upper()}"):
            # i of enumerate is needed to decide replace (=0) or append (>0)

            dict_kegg_genes = self.get_kegg_gene_identifiers(kegg_species_identifier)
            gene_tag = f"{kegg_species_identifier}:"

            url_pathway_list = f"http://rest.kegg.jp/list/pathway/{kegg_species_identifier}"
            df_pathways = pd.read_csv(url_pathway_list, sep="\t", names=["path_id", "path_desc"])
            df_pathways[["path_name", "organism"]] = df_pathways.path_desc.str.split(pat=" - ", n=1, expand=True)
            # df_pathways.path_id = df_pathways.path_id.str.lstrip(
            #     kegg_species_identifier
            # )
            df_pathways.drop(columns=["path_desc"], inplace=True)

            desc = f"Inserting KEGG data for KEGG species: {kegg_species_identifier}"

            for p in tqdm(df_pathways.to_dict("records"), desc=desc):
                xml_content = self._get_pathway_kgml(p["path_id"])
                dict_entries = self.get_entries(xml_content)
                relations = self.get_relations(xml_content)

                # Currently it only take interactions between proteins/genes. Ignores compounds and links to paths
                generic_table_rows = []
                for relation in relations:
                    entry_a, entry_b = (
                        dict_entries[relation["entry1"]],
                        dict_entries[relation["entry2"]],
                    )
                    kegg_int_type = relation["kegg_int_type"]
                    interaction_type = relation["interaction_type"]

                    # TODO implement compound
                    for kegg_gene_a in entry_a[0]:
                        for kegg_gene_b in entry_b[0]:
                            if gene_tag in kegg_gene_b and gene_tag in kegg_gene_a:
                                gene_symbol_a = dict_kegg_genes.get(kegg_gene_a)
                                gene_symbol_b = dict_kegg_genes.get(kegg_gene_b)
                                if gene_symbol_a and gene_symbol_b:
                                    generic_table_rows.append(
                                        [
                                            kegg_species_identifier,
                                            kegg_gene_a,
                                            gene_symbol_a,
                                            kegg_gene_b,
                                            gene_symbol_b,
                                            interaction_type,
                                            kegg_int_type,
                                            p["path_id"],
                                            p["path_name"],
                                        ]
                                    )

                df = pd.DataFrame(
                    generic_table_rows,
                    columns=[
                        "kegg_species_id",
                        "kegg_gene_id_a",
                        "gene_symbol_a",
                        "kegg_gene_id_b",
                        "gene_symbol_b",
                        "interaction_type",
                        "kegg_int_type",
                        "pathway_identifier",
                        "pathway_name",
                    ],
                )

                dfs.append(df)

        df_all = pd.concat(dfs)
        df_all.explode("interaction_type").to_sql(
            kegg.Kegg.__tablename__, self.engine, if_exists="append", index=False
        )

        return {self.biodb_name: df_all.shape[0]}

    def get_pure_rid(self, class_name, symbol, all_prot_symbol_rid_dict):
        """Get pure protein node rID with the metadata."""
        rid = None
        bel_class_mapper = {"protein": "p", "rna": "r"}
        if symbol in all_prot_symbol_rid_dict:
            rid = all_prot_symbol_rid_dict[symbol]
        else:
            correct_symbol = self.hgnc.get_correct_symbol(symbol)
            if correct_symbol:
                short_bel_function = bel_class_mapper[class_name]
                val_dict = {
                    "name": correct_symbol,
                    "namespace": "HGNC",
                    "bel": f'{short_bel_function}(HGNC:"{correct_symbol}")',
                    "pure": True,
                }
                rid = self.get_create_rid(class_name, val_dict, check_for="bel")
                all_prot_symbol_rid_dict[correct_symbol] = rid
        return rid

    def get_create_pure_protein(self, namespace: str, name: str) -> str:
        """Create or get pure protein node with the given namespace amd value."""
        bel = f'p({namespace}:"{name}")'
        if bel in self.bel_rid_cache:
            rid = self.bel_rid_cache[bel]
        else:
            value_dict = {
                "bel": bel,
                "pure": True,
                "namespace": namespace,
                "name": name,
            }
            rid = self.get_create_rid("protein", value_dict=value_dict, check_for="bel")
            self.bel_rid_cache[bel] = rid
        return rid

    def update_interactions(self) -> int:
        """Update edges with KEGG metadata."""
        self.clear_edges()

        symbol_rids_dict = self.get_pure_symbol_rids_dict_in_bel_context()

        inserted = 0

        pmods = {
            "dephosphorylation": ("pho", "decreases", BelPmod.PHO),
            "glycosylation": ("gly", "increases", BelPmod.GLY),
            "methylation": ("me0", "increases", BelPmod.ME0),
            "phosphorylation": ("pho", "increases", BelPmod.PHO),
            "ubiquitination": ("ubi", "increases", BelPmod.UBI),
        }
        post_translational_modifications = ",".join([f"'{x}'" for x in pmods.keys()])

        species_ids = ",".join([f"'{x}'" for x in self.species])

        sql_temp = f"""Select
                interaction_type,
                pathway_identifier,
                pathway_name,
                gene_symbol_a,
                gene_symbol_b,
                kegg_species_id
            from
                kegg
            where
                (gene_symbol_a='{{symbol}}' or gene_symbol_a='{{symbol}}') and
                kegg_species_id in ({species_ids}) and
                interaction_type in ({{interaction_types}})
            group by
                interaction_type,
                pathway_identifier,
                pathway_name,
                gene_symbol_a,
                gene_symbol_b,
                kegg_species_id"""

        for symbol, rid in tqdm(symbol_rids_dict.items(), desc="Update KEGG posttranslational modifications"):
            sql = sql_temp.format(symbol=symbol, interaction_types=post_translational_modifications)
            df = pd.read_sql(sql, self.engine)
            keys = (
                "interaction_type",
                "gene_symbol_a",
                "gene_symbol_b",
                "kegg_species_id",
            )

            grouped_records = dict()
            for _, row in df.iterrows():
                index = row[[keys]]
                pname = row["pathway_name"]

                if index in grouped_records:
                    grouped_records[index].add(pname)

                else:
                    grouped_records[index] = {pname}

            # TODO not working so using the more verbose method above
            # grouped_records = df.groupby(by=by).apply(lambda x: set(x.pathway_name)).to_dict()

            for key, pathway_names in grouped_records.items():
                interaction_type, subject_name, object_name, kegg_species_id = key

                namespace = self.species_namespace_dict[kegg_species_id]

                subject_rid = self.get_create_pure_protein(namespace, subject_name)
                object_rid = self.get_create_pure_protein(namespace, object_name)

                ebel_pmod, effect, bel_pmod = pmods[interaction_type]

                pmod_bel = f'p({namespace}:"{object_name}",pmod({bel_pmod}))'
                pmod_value_dict = {
                    "name": object_name,
                    "namespace": namespace,
                    "bel": pmod_bel,
                }

                pmod_protein_rid = self.node_exists("protein", pmod_value_dict, check_for="bel")

                if not pmod_protein_rid:
                    pmod_protein_rid = self.get_create_rid("protein", pmod_value_dict, check_for="bel")
                    self.create_edge("has_modified_protein", object_rid, pmod_protein_rid)
                    pmod_rid = self.insert_record("pmod", {"bel": f"pmod({bel_pmod})", "type": ebel_pmod})
                    self.create_edge("has__pmod", pmod_protein_rid, pmod_rid)

                edge_class = f"{effect}_{ebel_pmod}_kg"
                edge_value_dict = {
                    "interaction_type": interaction_type,
                    "pathway_name": list(pathway_names),
                }
                self.create_edge(edge_class, subject_rid, pmod_protein_rid, edge_value_dict)
                inserted += 1

        self.hgnc.update_bel()

        return inserted



"""STRING module."""
import logging
from collections import namedtuple
from typing import Dict

import numpy as np
import pandas as pd
from pyorientdb import OrientDB
from tqdm import tqdm

from ebel.manager.orientdb import odb_meta, odb_structure, urls
from ebel.manager.orientdb.biodbs.hgnc import Hgnc
from ebel.manager.orientdb.constants import STRINGDB
from ebel.manager.rdbms.models import stringdb
from ebel.tools import get_file_path

logger = logging.getLogger(__name__)


class StringDb(odb_meta.Graph):
    """STRING database."""

    def __init__(self, client: OrientDB = None):
        """Init StringDB database."""
        self.client = client
        self.biodb_name = STRINGDB
        self.table_action = stringdb.StringDbAction.__tablename__
        self.table_strdb = stringdb.StringDb.__tablename__
        self.table_protein = stringdb.StringDbProtein.__tablename__
        self.urls = {
            self.table_strdb: urls.STRING_INTS,
            self.table_action: urls.STRING_ACTIONS,
            self.table_protein: urls.STRING_NAMES,
        }
        super().__init__(
            edges=odb_structure.stringdb_edges,
            urls=self.urls,
            tables_base=stringdb.Base,
            biodb_name=self.biodb_name,
        )

    def __len__(self) -> dict:
        """Get number of 'biogrid_interaction' graph edges."""
        pass

    def __contains__(self, biogrid_id) -> bool:
        """Check if stringdb edge in graph."""
        pass

    def insert_data(self) -> Dict[str, int]:
        """Insert data from url. Return number of inserted entries."""
        inserted = dict()
        inserted[self.table_protein] = self.insert_protein_data()
        inserted[self.table_strdb] = self.insert_link_data()
        inserted[self.table_action] = self.insert_action_data()

        self.session.commit()
        return inserted

    def insert_link_data(self) -> int:
        """Insert link STRINGDB information into RDBMS."""
        logger.info(f"Insert {self.biodb_name} link data in RDBMS.")

        file_path = get_file_path(self.urls[self.table_protein], self.biodb_name)
        df_protein = pd.read_csv(file_path, sep="\t", usecols=["#string_protein_id", "preferred_name"]).rename(
            columns={"#string_protein_id": "string_protein_id"}
        )

        # Define column types to improve memory efficiency
        cols = [
            "protein1",
            "protein2",
            "neighborhood",
            "neighborhood_transferred",
            "fusion",
            "cooccurence",
            "homology",
            "coexpression",
            "coexpression_transferred",
            "experiments",
            "experiments_transferred",
            "database",
            "database_transferred",
            "textmining",
            "textmining_transferred",
            "combined_score",
        ]

        col_types = dict()
        for col in cols:
            if col.startswith("protein"):
                col_types[col] = "object"
            else:
                col_types[col] = "uint16"

        file_path = get_file_path(self.urls[self.table_strdb], self.biodb_name)
        df_link = pd.read_csv(file_path, dtype=col_types, sep=" ")

        df = df_link.set_index("protein1").join(df_protein.set_index("string_protein_id"), how="inner")
        df.reset_index(inplace=True)
        df.rename(columns={"index": "protein1", "preferred_name": "symbol1"}, inplace=True)
        df = df.set_index("protein2").join(df_protein.set_index("string_protein_id"), how="inner")
        df.reset_index(inplace=True)
        df.rename(columns={"index": "protein2", "preferred_name": "symbol2"}, inplace=True)
        df.index += 1
        df.index.rename("id", inplace=True)

        df.to_sql(self.table_strdb, self.engine, if_exists="append", chunksize=10000)

        return df.shape[0]

    def insert_protein_data(self) -> int:
        """Generates a dictionary of STRINGDB identifiers as keys and HGNC symbols as values."""
        file_path = get_file_path(self.urls[self.table_protein], self.biodb_name)
        df = pd.read_csv(file_path, sep="\t", usecols=["#string_protein_id", "preferred_name"]).rename(
            columns={"#string_protein_id": "string_protein_id"}
        )
        df.index += 1
        df.index.rename("id", inplace=True)
        df.to_sql(self.table_protein, self.engine, if_exists="append")
        return df.shape[0]

    def insert_action_data(self) -> int:
        """Insert action STRINGDB information into RDGMS."""
        logger.info(f"Insert {self.biodb_name} action data in RDBMS.")

        file_path = get_file_path(self.urls[self.table_protein], self.biodb_name)
        df_protein = pd.read_csv(file_path, sep="\t", usecols=["#string_protein_id", "preferred_name"]).rename(
            columns={"#string_protein_id": "string_protein_id"}
        )

        file_path = get_file_path(self.urls[self.table_action], self.biodb_name)
        df_action = pd.read_csv(file_path, sep="\t")

        # Replace 't' and 'f' values for True/False in appropriate columns
        df_action.replace(
            {
                "is_directional": {"t": True, "f": False},
                "a_is_acting": {"t": True, "f": False},
            },
            inplace=True,
        )

        # Replace NaN with None
        df_action.replace({np.nan: None}, inplace=True)

        df = df_action.set_index("item_id_a").join(df_protein.set_index("string_protein_id"), how="inner")
        df.reset_index(inplace=True)
        df.rename(columns={"index": "item_id_a", "preferred_name": "symbol1"}, inplace=True)
        df = df.set_index("item_id_b").join(df_protein.set_index("string_protein_id"), how="inner")
        df.reset_index(inplace=True)
        df.rename(columns={"index": "item_id_b", "preferred_name": "symbol2"}, inplace=True)
        df.index += 1
        df.index.rename("id", inplace=True)

        df.to_sql(self.table_action, self.engine, if_exists="append", chunksize=10000)

        return df.shape[0]

    def get_stringdb_action_hgnc_set(self):
        """Get unique HGNC symbols from stringdb_actions table."""
        sql = f"""(Select distinct( symbol1 ) from {self.table_action})
                union (Select distinct( symbol2 ) from {self.table_action})"""
        return set([x[0] for x in self.engine.execute(sql).fetchall()])

    def update_interactions(self) -> Dict[str, int]:
        """Update the edges with StringDB metadata."""
        hgnc = Hgnc(self.client)
        updated = dict()
        updated["interactions"] = self.update_stringdb_interactions(hgnc)
        updated["actions"] = self.update_action_interactions(hgnc)
        return updated

    def get_stringdb_symbols(self):
        """Return all gene symbols used by StringDB."""
        return {x[0] for x in self.session.query(stringdb.StringDbProtein.preferred_name).all()}

    def update_stringdb_interactions(self, hgnc: Hgnc) -> int:
        """Iterate through BEL proteins and adds stringdb edges to existing proteins in KG."""
        columns = (
            "neighborhood",
            "neighborhood_transferred",
            "fusion",
            "cooccurence",
            "homology",
            "coexpression",
            "coexpression_transferred",
            "experiments",
            "experiments_transferred",
            "database",
            "database_transferred",
            "textmining",
            "textmining_transferred",
            "combined_score",
        )

        bel_hgnc_rid_dict = self.get_pure_symbol_rids_dict_in_bel_context(namespace="HGNC")
        bel_hgncs = set(bel_hgnc_rid_dict.keys())
        strdb_hgncs = self.get_stringdb_symbols()
        shared_hgncs = bel_hgncs & strdb_hgncs

        updated = 0
        already_inserted = set()

        strdb = stringdb.StringDb

        for hgnc_a in tqdm(shared_hgncs, desc="Update has_ppi_st edges"):
            query = self.session.query(strdb)
            query = query.filter(strdb.symbol1 == hgnc_a, strdb.experiments > 0)

            for row in query.all():
                sorted_combi = tuple(sorted([row.symbol1, row.symbol2]))

                if sorted_combi not in already_inserted:
                    value_dict = {k: v for k, v in row.__dict__.items() if k in columns}

                    from_rid = self.get_create_rid_by_symbol(row.symbol1, bel_hgnc_rid_dict, hgnc)
                    to_rid = self.get_create_rid_by_symbol(row.symbol2, bel_hgnc_rid_dict, hgnc)

                    if from_rid and to_rid:
                        self.create_edge(
                            class_name="has_ppi_st",
                            from_rid=from_rid,
                            to_rid=to_rid,
                            value_dict=value_dict,
                        )
                        already_inserted.add(sorted_combi)
                        updated += 1

        return updated

    def get_create_rid_by_symbol(self, symbol: str, symbol_rid_dict: dict, hgnc: Hgnc) -> str:
        """Create or get rID entry for a given gene symbol.

        Parameters
        ----------
        symbol: str
            Gene symbol.
        symbol_rid_dict: dict
            Entry parameters matching those of the desired rID entry.
        hgnc: Hgnc
            Hgnc model definition.

        Returns
        -------
        str
            rID.
        """
        if symbol not in symbol_rid_dict:
            symbol = hgnc.get_correct_symbol(symbol)
            if symbol:
                value_dict = {
                    "name": symbol,
                    "namespace": "HGNC",
                    "pure": True,
                    "bel": f'p(HGNC:"{symbol}")',
                }
                symbol_rid_dict[symbol] = self.get_create_rid("protein", value_dict, check_for="bel")
        return symbol_rid_dict.get(symbol)

    def update_action_interactions(self, hgnc: Hgnc) -> int:
        """Iterate through BEL proteins and add stringdb_action edges to existing proteins in KG.

        Translation of String action/mode combination to ebel-edges

        mode       | action     | ebel edge class
        -----------|------------|-----------
        expression | NULL       | increases-expression-of
        ptmod      | NULL       | controls-pmod-of
        activation | activation | activates
        expression | inhibition | decreases-expression-of
        inhibition | inhibition | inhibits
        """
        translator = {
            ("expression", None): "increases_expression_of_st",
            ("ptmod", None): "controls_pmod_of_st",
            ("activation", "activation"): "activates_st",
            ("expression", "inhibition"): "decreases_expression_of_st",
            ("inhibition", "inhibition"): "inhibits_st",
        }

        Action = namedtuple("Action", ("symbol1", "symbol2", "mode", "action", "score"))

        columns = ", ".join(Action._fields)
        sql_temp = f"""Select {columns} from {self.table_action}
                       where mode in ('activation', 'inhibition', 'ptmod', 'expression')
                       and (symbol1='{{symbol}}' or symbol2='{{symbol}}')
                       and is_directional=1 and a_is_acting=1"""

        symbols_rid_dict = self.get_pure_symbol_rids_dict_in_bel_context(namespace="HGNC")
        symbols = tuple(symbols_rid_dict.keys())

        already_inserted = set()

        updated = 0
        for symbol in tqdm(symbols, desc="Update has_action_st edges"):
            rows = self.engine.execute(sql_temp.format(symbol=symbol))
            for row in rows.fetchall():
                action = Action(*row)

                sorted_combi = tuple(sorted([action.symbol1, action.symbol2]))

                if sorted_combi not in already_inserted:
                    from_rid = self.get_create_rid_by_symbol(action.symbol1, symbols_rid_dict, hgnc)
                    to_rid = self.get_create_rid_by_symbol(action.symbol2, symbols_rid_dict, hgnc)

                    if from_rid and to_rid:
                        class_name = translator[(action.mode, action.action)]

                        self.create_edge(
                            class_name=class_name,
                            from_rid=from_rid,
                            to_rid=to_rid,
                            value_dict={"score": action.score},
                        )
                        already_inserted.update([sorted_combi])
                        updated += 1

        return updated



"""Methods for handling the configuration file."""
import configparser
import logging
import os
import random
import re
import string
from collections import namedtuple
from configparser import RawConfigParser
from getpass import getpass
from typing import Optional, Union
from urllib.parse import quote

import pymysql

from ebel import defaults
from ebel.constants import DEFAULT_ODB
from ebel.constants import TerminalFormatting as TF
from ebel.defaults import CONN_STR_DEFAULT, DATABASE_LOCATION


def set_configuration(
    name: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    server: Optional[str] = None,
    port: Union[str, int, None] = None,
    user_reader: Optional[str] = None,
    user_reader_password: Optional[str] = None,
    root_password: Optional[str] = None,
    kegg_species: Optional[str] = None,
    sqlalchemy_connection_string: Optional[str] = None,
    snp_related_traits: Optional[str] = None,
    drugbank_user: Optional[str] = None,
    drugbank_password: Optional[str] = None,
) -> dict:
    """Set configuration values in the config file."""
    odb_class_attribs = {
        "name": name,
        "user": user,
        "password": password,
        "server": server,
        "port": port,
        "user_reader": user_reader,
        "user_reader_password": user_reader_password,
        "root_password": root_password,
    }

    for param, value in odb_class_attribs.items():
        write_to_config(DEFAULT_ODB, param, value)

    if kegg_species:
        write_to_config("KEGG", "species", kegg_species)

    if sqlalchemy_connection_string:
        write_to_config("DATABASE", "sqlalchemy_connection_string", sqlalchemy_connection_string)

    if snp_related_traits:
        write_to_config("SNP_RELATED_TRAITS", "keyword", snp_related_traits)

    if drugbank_user:
        write_to_config("DRUGBANK", "user", drugbank_user)

    if drugbank_password:
        write_to_config("DRUGBANK", "password", drugbank_password)

    return get_config_as_dict()


def write_to_config(section: str, option: str, value: str) -> None:
    """Write section, option and value to config file.

    Parameters
    ----------
    section : str
        Section name of configuration file.
    option : str
        Option name.
    value : str
        Option value.
    """
    if value:
        cfp = defaults.config_file_path
        config = RawConfigParser()

        if not os.path.exists(cfp):
            with open(cfp, "w") as config_file:
                config[section] = {option: value}
                config.write(config_file)
                logging.info(f"Set in configuration file {cfp} in section {section} {option}={value}")
        else:
            config.read(cfp)
            if not config.has_section(section):
                config.add_section(section)
            config.set(section, option, value)
            with open(cfp, "w") as configfile:
                config.write(configfile)


def get_config_as_dict():
    """Get ebel configuration as dictionary if config file exists."""
    cfp = defaults.config_file_path
    if os.path.exists(cfp):
        config = RawConfigParser()
        config.read(cfp)
        return config._sections

    else:
        return user_config_setup(config_exists=False)


def user_config_setup(config_exists: bool = True) -> dict:
    """Set up the configuration file."""
    old_configs = get_config_as_dict() if config_exists else {}
    configs = {}

    print(f"\n{TF.TITLE}e(BE:L) Configuration WIZARD{TF.RESET}")
    print("\nThe following questionnaire will guide you through the configuration process.\n")
    print(
        "Before we start: Make sure\n\t1. OrientDB and\n\t2. MySQL/MariaDB (optional)\n are running and "
        "you have the root password for both, if databases and users do not already exist.\n"
    )
    print(
        f"Default {TF.DEFAULT_VALUE}[values]{TF.RESET} are written in square brackets and "
        f"can be confirmed by RETURN.\n"
    )

    print(f"{TF.Format.UNDERLINED}Installation references{TF.RESET}")
    print(f"\t OrientDB: {TF.Fore.BLUE}https://orientdb.org/docs/3.1.x/fiveminute/java.html{TF.RESET}")
    print(
        f"\t MySQL: {TF.Fore.BLUE}"
        f"https://dev.mysql.com/doc/mysql-getting-started/en/#mysql-getting-started-installing{TF.RESET}"
    )
    print(f"\t MariaDB: {TF.Fore.BLUE}https://mariadb.com/kb/en/getting-installing-and-upgrading-mariadb/{TF.RESET}\n")

    # RDBMS setup
    old_sa_con_str = old_configs.get("sqlalchemy_connection_string", "").strip()
    configs["sqlalchemy_connection_string"] = __user_rdbms_setup(old_sa_con_str)

    # OrientDB Setup
    print(f"\n{TF.HEADER}Graph database (OrientDB) settings{TF.RESET}")

    Configuration = namedtuple(
        "Configuration",
        ["name", "question", "default", "validation_regex", "is_password", "required"],
    )

    old_configs_odb = old_configs.get(DEFAULT_ODB, {})
    orientdb_configs = {
        "name": (
            "OrientDB database name (created if not exists)",
            old_configs_odb.get("name", "ebel"),
            r"^[A-Za-z]\w{2,}$",
            False,
            True,
        ),
        "user": (
            "OrientDB user (admin) name (created if not exists)",
            old_configs_odb.get("user", "ebel_user"),
            r"^[A-Za-z]\w{2,}$",
            False,
            True,
        ),
        "password": (
            "OrientDB user (admin) password (created if not exists)",
            "".join(random.sample(string.ascii_letters, 12)),
            None,
            True,
            True,
        ),
        "server": (
            "OrientDB server",
            old_configs_odb.get("server", "localhost"),
            None,
            False,
            True,
        ),
        "port": (
            "OrientDB port",
            old_configs_odb.get("port", "2424"),
            r"^\d+$",
            False,
            True,
        ),
        "user_reader": (
            "OrientDB user (reader) name (created if not exists).",
            old_configs_odb.get("user_reader", "ebel_reader"),
            r"^[A-Za-z]\w{2,}$",
            False,
            False,
        ),
        "user_reader_password": (
            "OrientDB user (reader) password (created if not exists).",
            "".join(random.sample(string.ascii_letters, 12)),
            None,
            True,
            True,
        ),
    }

    for param_name, options in orientdb_configs.items():
        conf = Configuration(param_name, *options)
        input_method = getpass if conf.is_password else input
        invalid_value = True
        while invalid_value:
            if conf.default:
                question = f"{TF.QUESTION}{conf.question}{TF.RESET} {TF.DEFAULT_VALUE}[{conf.default}]{TF.RESET}"
                configs[conf.name] = input_method(f"{question}: ").strip() or conf.default

            else:
                configs[conf.name] = input_method(f"{TF.QUESTION}{conf.question}{TF.RESET}").strip()

            if conf.validation_regex and conf.required:
                invalid_value = not bool(re.search(conf.validation_regex, configs[conf.name]))
                if invalid_value:
                    print("!!!>>> WARNING <<<!!!\nInvalid entry. Please select a valid option")

            else:
                invalid_value = False

    # KEGG
    print(f"\n{TF.HEADER}KEGG settings{TF.RESET}")

    kegg_question = (
        f"{TF.QUESTION}KEGG species as 3-4 letter code comma separated.{TF.RESET} "
        "(see here for the KEGG organism letter codes: "
        f"{TF.Fore.BLUE}https://www.genome.jp/kegg/catalog/org_list4.html{TF.RESET} ) "
        f"{TF.DEFAULT_VALUE}[hsa,rno,mmu]: {TF.RESET}"
    )
    configs["kegg_species"] = input(kegg_question).strip() or "hsa,rno,mmu"

    print(f"\n{TF.HEADER}SNP related traits settings{TF.RESET}")

    # SNP
    default_snp_related_traits = old_configs.get("snp_related_traits") or "Alzheimer,Parkinson"
    snp_related_traits_question = (
        f"{TF.QUESTION}SNPs related to (separated by comma){TF.RESET} "
        f"{TF.DEFAULT_VALUE}[{default_snp_related_traits}]: {TF.RESET}"
    )
    snp_related_traits = input(snp_related_traits_question).strip()
    configs["snp_related_traits"] = snp_related_traits or default_snp_related_traits

    # DrugBank
    print(f"\n{TF.HEADER}DrugBank settings{TF.RESET}")
    print("If you have a valid DrugBank account (otherwise leave it blank): ")
    drugbank_user = input(f"{TF.QUESTION}DrugBank user: {TF.RESET}").strip()

    if drugbank_user:
        configs["drugbank_user"] = drugbank_user
        configs["drugbank_password"] = getpass(f"{TF.QUESTION}DrugBank password: {TF.RESET}")

    current_config = set_configuration(**configs)

    return current_config


def __user_orientdb_setup():
    """Set up ODB params."""
    pass


def __user_rdbms_setup(prev_conn: str) -> str:
    """The initial setup process for the RDBMS."""
    db_choice = (
        input(
            """e(BE:L) requires some basic information in order to begin importing data and building a
Knowledge Graph. The nodes and edges compiled from BEL statements are imported into OrientDB while the information
parsed from external repositories is stored in a more traditional relational database and uses either SQLite or MySQL.

While SQLite is easier to set up and does not require installing additional software, MySQL is the recommended
option due to the amount of information that will be imported.

MySQL/SQLite [MySQL]: """
        )
        or "mysql"
    )

    while db_choice.lower() not in ("sqlite", "mysql"):
        db_choice = input("Bad input, please enter either 'MySQL' or 'SQLite': ")

    if db_choice.lower() == "sqlite":
        db_conn = CONN_STR_DEFAULT
        print(f"SQLite DB can be found at {DATABASE_LOCATION}")

    else:  # MySQL
        db_conn = __mysql_setup(prev_conn)

    print("Database set. Connection string can be changed using `ebel set_connection` or `ebel set_mysql`")
    return db_conn


def __mysql_setup(old_sa_con_str: str) -> str:
    """Set up the MySQL/MariaDB connection and return the conenction string."""
    print(f"\n{TF.HEADER}MySQL/MariaDB settings{TF.RESET}")

    old_mysql = {}

    if old_sa_con_str:
        regex_con_str = (
            r"^mysql\+pymysql://(?P<mysql_user>.*?):" r"(?P<mysql_passwd>.*?)@(?P<mysql_host>.*?)/(?P<mysql_db>.*)$"
        )
        found_old_mysql = re.search(regex_con_str, old_sa_con_str)
        if found_old_mysql:
            old_mysql = found_old_mysql.groupdict()

    default_mysql_host = old_mysql.get("mysql_host") or "localhost"
    mysql_host_question = (
        f"{TF.QUESTION}MySQL/MariaDB server name{TF.RESET} " f"{TF.DEFAULT_VALUE}[{default_mysql_host}]: {TF.RESET}"
    )
    mysql_host = input(mysql_host_question) or default_mysql_host

    default_mysql_port = old_mysql.get("mysql_db") or "3306"
    mysql_port_question = (
        f"{TF.QUESTION}MySQL/MariaDB port{TF.RESET} " f"{TF.DEFAULT_VALUE}[{default_mysql_port}]: {TF.RESET}"
    )
    mysql_port = input(mysql_port_question).strip() or default_mysql_port

    default_mysql_user = old_mysql.get("mysql_user") or "ebel"
    mysql_user_question = (
        f"{TF.QUESTION}MySQL/MariaDB (non-root) user{TF.RESET} "
        f"{TF.DEFAULT_VALUE}[{default_mysql_user}]: {TF.RESET}"
    )
    mysql_user = input(mysql_user_question).strip() or default_mysql_user

    mysql_random_password = "".join(random.sample(string.ascii_letters, 12))
    mysql_passed_question = (
        f"{TF.QUESTION}MySQL/MariaDB password for user{TF.RESET} "
        f"{TF.DEFAULT_VALUE}[{mysql_random_password}]: {TF.RESET}"
    )
    mysql_pwd = getpass(mysql_passed_question).strip() or mysql_random_password

    default_mysql_db = old_mysql.get("mysql_db") or "ebel"
    mysql_db_question = (
        f"{TF.QUESTION}MySQL/MariaDB database name{TF.RESET} " f"{TF.DEFAULT_VALUE}[{default_mysql_db}]: {TF.RESET}"
    )
    mysql_db = input(mysql_db_question).strip() or default_mysql_db

    db_conn = f"mysql+pymysql://{mysql_user}:{quote(mysql_pwd)}@{mysql_host}:{mysql_port}/{mysql_db}?charset=utf8mb4"

    try:
        pymysql.connect(host=mysql_host, user=mysql_user, password=mysql_pwd, db=mysql_db)

    except pymysql.err.OperationalError:
        mysql_root_passwd = getpass(
            f"{TF.QUESTION}MySQL root password (will be not stored) " f"to create database and user: {TF.RESET}"
        )
        print(mysql_host, "root", mysql_root_passwd)
        cursor = pymysql.connect(host=mysql_host, user="root", password=mysql_root_passwd).cursor()
        db_exists = cursor.execute("show databases like %s", mysql_db)
        if not db_exists:
            cursor.execute(f"CREATE DATABASE {mysql_db} CHARACTER SET utf8mb4 COLLATE utf8mb4_general_ci")
        else:
            print(f"Database {mysql_db} already exists.")

        user_exists = cursor.execute("Select 1 from mysql.user where User=%s", mysql_user)

        if user_exists:
            print("USer already exists, will be recreated")
            cursor.execute(f"DROP USER '{mysql_user}'@'%'")
            cursor.execute("FLUSH PRIVILEGES")
        sql = f"CREATE USER '{mysql_user}'@'%' IDENTIFIED BY '{mysql_pwd}'"
        print(sql)
        cursor.execute(sql)
        cursor.execute("FLUSH PRIVILEGES")

        privileges_exists = cursor.execute("Select 1 from mysql.db where User=%s and Db=%s", (mysql_user, mysql_db))

        if not privileges_exists:
            sql = f"GRANT ALL PRIVILEGES ON `{mysql_db}`.*  TO '{mysql_user}'@'%'"
            print(sql)
            cursor.execute(sql)
            cursor.execute("FLUSH PRIVILEGES")

        else:
            print(f"MySQL user {mysql_user} already has sufficient rights to {mysql_db}")

    return db_conn


def get_config_value(section, option, value=None) -> Optional[str]:
    """Retrieve value from a given section and option from the config file if it exists.

    Parameters
    ----------
    section (str): Configuration section header
    option (str): Option value within the section
    value (str): (optional) value to set the option equal to if option doesn't exist

    Returns
    -------
    config_value: The value of the specified section and option
    """
    cfg = configparser.ConfigParser()

    if os.path.isfile(defaults.config_file_path):
        cfg.read(defaults.config_file_path)

        if cfg.has_section(section) and cfg.has_option(section=section, option=option):
            config_value = cfg[section][option]

        elif value is not None:
            write_to_config(section, option, value)
            config_value = value

        else:
            return None

    else:
        write_to_config(section, option, value)
        config_value = value

    return config_value



"""IntAct API methods."""
from flask.globals import request
from sqlalchemy import or_

from ebel.manager.orientdb.odb_structure import intact_edges
from ebel.manager.rdbms.models.intact import Intact
from ebel.web.api import RDBMS
from ebel.web.api.ebel.v1 import (_get_data, _get_paginated_ebel_query_result,
                                  _get_paginated_query_result)


def get_intact():
    """Get generic IntAct entry."""
    return _get_data(Intact)


def get_by_uniprot():
    """Get IntAct entry by UniProt ID."""
    ua = request.args.get("uniprot_accession")
    if ua:
        a = Intact.int_a_uniprot_id
        b = Intact.int_a_uniprot_id
        query = RDBMS.get_session().query(Intact).filter(or_(a == ua, b == ua))
        return _get_paginated_query_result(query)


def get_ebel_relation():
    """Get IntAct related eBEL relations."""
    has_ppi_ia_edge = [x for x in intact_edges if x.name == "has_ppi_ia"][0]
    conf = {x.prop_name: x.prop_name for x in has_ppi_ia_edge.props}
    conf.update(
        {
            "relation_type": "@class",
            "edge_id": "@rid.asString()",
            "interactor_a_rid": "out.@rid.asString()",
            "interactor_a_name": "out.name",
            "interactor_a_namespace": "out.namespace",
            "interactor_a_bel": "out.bel",
            "interactor_b_rid": "in.@rid.asString()",
            "interactor_b_namespace": "in.namespace",
            "interactor_b_name": "in.name",
            "interactor_b_bel": "in.bel",
        }
    )
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_ppi_ia"

    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + " AND ".join([f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()])

    return _get_paginated_ebel_query_result(sql)



"""miRTarBase API methods."""

from flask import request

from ebel.manager.rdbms.models import mirtarbase
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_mirtarbase():
    """Get generic mirtarbase entry."""
    return _get_data(mirtarbase.Mirtarbase)


def get_ebel_relation():
    """Get miRTarBase related eBEL relations."""
    conf = {
        "mirbase_id": "out.name",
        "target_rna_symbol": "in.name",
        "target_namespace": "in.namespace",
        "support_type": "support_type",
        "pmid": "pmid",
        "experiments": "experiments",
    }
    sql = "SELECT "
    sql += ", ".join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_mirgene_target"

    ra = request.args

    wheres = []

    experiment = ra.get("experiment")
    if experiment:
        wheres.append(f'"{experiment}" in experiments')

    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]

    if wheres:
        sql += " WHERE " + " AND ".join(wheres)

    return _get_paginated_ebel_query_result(sql)
